from . import (
    CF,
    AuxiliaryCoordinate,
    Bounds,
    CellConnectivity,
    CellMeasure,
    CellMethod,
    CoordinateConversion,
    CoordinateReference,
    Count,
    Datum,
    DimensionCoordinate,
    Domain,
    DomainAncillary,
    DomainAxis,
    DomainTopology,
    Field,
    FieldAncillary,
    Index,
    InteriorRing,
    InterpolationParameter,
    List,
    NodeCountProperties,
    PartNodeCountProperties,
    TiePointIndex,
)
from .abstract import Implementation
from .data import (
    AggregatedArray,
    BoundsFromNodesArray,
    CellConnectivityArray,
    Data,
    GatheredArray,
    H5netcdfArray,
    NetCDF4Array,
    PointTopologyArray,
    RaggedContiguousArray,
    RaggedIndexedArray,
    RaggedIndexedContiguousArray,
    SubsampledArray,
)


class CFDMImplementation(Implementation):
    """A container for the CF data model implementation.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return f"<{self.__class__.__name__}: >"

    def _get_domain_compression_variable(self, variable_type, domain):
        """Get the compression variable of a type of compressed data.

        ..versionadded:: (cfdm) 1.9.0.0

        :Parameters:

            variable_type: `str`

            domain: Domain construct

        :Returns:

            Compression variable or `None`

        """
        if variable_type == "count":
            compression_types = (
                "ragged contiguous",
                "ragged indexed contiguous",
            )
        elif variable_type == "index":
            compression_types = ("ragged indexed", "ragged indexed contiguous")
        elif variable_type == "list":
            compression_types = ("gathered",)
        else:
            raise ValueError(
                f"Invalid value for 'variable_type': {variable_type!r}"
            )

        constructs = self.get_constructs(domain, data=True)

        variable = None
        for key, c in constructs.items():
            compression_type = self.get_compression_type(c)
            if compression_type not in compression_types:
                continue

            data = self.get_data(c, None)
            if data is None:
                continue

            if variable_type == "count":
                variable1 = data.get_count(None)
            elif variable_type == "index":
                variable1 = data.get_index(None)
            elif variable_type == "list":
                variable1 = data.get_list(None)

            if variable1 is None:
                continue

            if variable is not None and not self.equal_components(
                variable, variable1
            ):
                raise ValueError(
                    "Can't write domain variable from {!r} that is "
                    "associated with multiple count variables".format(domain)
                )

            variable = variable1

        return variable

    def bounds_insert_dimension(self, bounds, position):
        """Insert a new dimension into bounds data.

        :Parameters:

            bounds: bounds component

            position: `int`

        :Returns:

            Bounds component

        """
        return bounds.insert_dimension(position=position)

    def climatological_time_axes(self, construct):
        """Return all axes which are climatological time axes.

        .. versionadded:: (cfdm) 1.9.0.0

        :Parameters:

            construct: field or domain construct

        :Returns:

            `set`

        """
        return construct.climatological_time_axes()

    def conform_geometry_variables(self, field):
        """Collate geometry variable properties.

        .. versionadded:: (cfdm) 1.8.0

        :Parameters:

            field: field construct

        :Returns:

            `bool`

        """
        out = {"node_count": {}, "part_node_count": {}, "interior_ring": {}}

        for coord in self.get_auxiliary_coordinates(field).values():
            for variable in out:
                x = getattr(self, "get_" + variable)(coord)
                if x is None:
                    continue

                for k, v in x.properties().items():
                    if k not in out[variable]:
                        out[variable][k] = v
                    elif v != out[variable][k]:
                        return False

        for coord in self.get_auxiliary_coordinates(field).values():
            for variable in out:
                x = getattr(self, "get_" + variable)(coord)
                if x is None:
                    continue

                x.set_properties(out[variable])

        return True

    def construct_insert_dimension(self, construct, position):
        """Insert a new dimension into metadata construct data.

        :Parameters:

            construct: construct

            position: `int`

        :Returns:

            construct

        """
        return construct.insert_dimension(position=position)

    def copy_construct(self, construct):
        """Return a deep copy of a construct.

        :Returns:

                The deep copy.

        """
        return construct.copy()

    def convert(self, field=None, construct_id=None):
        """Convert a metadata construct into a field construct.

        Only metadata constructs that can have data may be converted
        and they can be converted even if they do not actually have
        any data. Constructs such as cell methods which cannot have
        data cannot be converted.

        :Parameters:

            field: field construct

            construct_id: `str`

        :Returns:

            Field construct

        """
        return field.convert(construct_id, full_domain=False)

    def data_insert_dimension(self, data, position):
        """Insert a new dimension into a data array.

        :Parameters:

            bounds: data

            position: `int`

        :Returns:

                The new data array

        """
        return data.insert_dimension(position=position)

    def del_properties(self, construct, props):
        """Remove a property from a construct.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            construct: construct

            props: (sequence of) `str`

            default: optional

        :Returns:

            `dict`
                The deleted properties, if any

        """
        return construct.del_properties(props)

    def del_property(self, construct, prop, default):
        """Remove a property from a construct.

        :Parameters:

            construct: construct

            prop: `str`

            default: optional

        :Returns:

                The value of the deleted property

        """
        return construct.del_property(prop, default)

    def field_insert_dimension(self, field, position=0, axis=None):
        """Insert a new dimension into field construct data.

        :Parameters:

            field: field construct

            position: `int`

            axis: optional

        :Returns:

            field construct

        """
        return field.insert_dimension(axis, position=position)

    def get_array(self, data):
        """Return the data as a `numpy` array.

        :Parameters:

            data: data instance

        :Returns:

            `numpy.ndarray`

        """
        return data.array

    def get_auxiliary_coordinates(self, field, axes=None, exact=False):
        """Returns auxiliary coordinates that span particular axes.

        If no axes are specified then all auxiliary coordinate constructs
        are returned.

        If axes are specified then auxiliary coordinate constructs whose
        data arrays span those axes, and possibly other axes, are
        returned.

        :Parameters:

            axes: sequence of `str`

        :Returns:

            `dict`

        """
        # To avoid mutable default argument (an anti-pattern) of axes=[]
        if axes is None:
            return field.auxiliary_coordinates(todict=True)

        if exact:
            axis_mode = "exact"
        else:
            axis_mode = "and"

        return field.auxiliary_coordinates(
            filter_by_axis=axes, axis_mode=axis_mode, todict=True
        )

    def get_bounds(self, parent, default=None):
        """Return the bounds of a construct.

        :Parameters:

            parent: construct with bounds

            default: optional

        :Returns:

            Bounds component

        """
        return parent.get_bounds(default=default)

    def get_bounds_ncvar(self, parent, default=None):
        """Return the netCDF variable name of the bounds.

        :Parameters:

            parent: construct with bounds

            default: optional

        :Returns:

            `str`

        """
        bounds = parent.get_bounds(None)
        if bounds is None:
            return default

        return self.nc_get_variable(bounds, default=default)

    def get_cell_measures(self, field):
        """Return all of the cell measure constructs of a field.

        :Parameters:

            field: field construct

        :Returns:

        """
        return field.cell_measures()

    def get_cell_methods(self, field):
        """Return all of the cell method constructs of a field.

        :Parameters:

            field: field construct

        :Returns:

        """
        return field.cell_methods()

    def get_cell_method_axes(self, cell_method, default=None):
        """Return the axes of a cell method construct.

        :Parameters:

            cell_method: cell method construct

            default: optional

        :Returns:

            `tuple`

        """
        return cell_method.get_axes(default=default)

    def get_cell_method_string(self, cell_method):
        """Return the `str` representation of a cell method construct.

        :Parameters:

            cell_method: cell method construct

        :Returns:

            `str`

        """
        return str(cell_method)

    def get_cell_method_qualifiers(self, cell_method):
        """Return the `str` representation of a cell method construct.

        :Parameters:

            cell_method: cell method construct

        :Returns:

            `str`

        """
        return cell_method.qualifiers()

    def get_compressed_array(self, data):
        """Return the data in its compressed form .

        :Parameters:

            data: data instance

        :Returns:

            `numpy.ndarray`

        """
        return data.compressed_array

    def get_compressed_axes(self, field_or_domain, key=None, construct=None):
        """Return the axes that have been compressed.

        :Parameters:

            field_or_domain: field or domain construct

            key: `str`, optional

            construct: optional
                If *construct* is set must *key* be set, too.

        :Returns:

            `list` of `str`
                The domain axis identifiers of dimensions that are
                compressed.

        """
        if construct is not None:
            data = self.get_data(construct)
            data_axes = self.get_construct_data_axes(field_or_domain, key)
            return [data_axes[i] for i in self.get_data_compressed_axes(data)]

        if self.is_field(field_or_domain):
            field = field_or_domain
            data = self.get_data(field)
            data_axes = self.get_field_data_axes(field)
            return [data_axes[i] for i in self.get_data_compressed_axes(data)]

        # For a domain construct, work out the compression axes from
        # its metadata constucts.
        domain = field_or_domain

        compression_types = (
            "gathered",
            "ragged indexed contiguous",
            "ragged indexed",
            "ragged contiguous",
            "subsampled",
        )

        compressed_axes = {
            compression_type: set() for compression_type in compression_types
        }

        constructs = self.get_constructs(domain, data=True)

        for key, c in constructs.items():
            compression_type = self.get_compression_type(c)
            if not compression_type:
                continue

            data_axes = self.get_construct_data_axes(domain, key)
            if not data_axes:
                continue

            data = self.get_data(c, None)
            if data is None:
                continue

            compressed_axes[compression_type].update(
                [data_axes[i] for i in self.get_data_compressed_axes(data)]
            )

        # The order of the following loop matters
        for compression_type in compression_types:
            if compressed_axes[compression_type]:
                return list(compressed_axes[compression_type])

        return []

    def get_compression_type(self, construct):
        """Returns the type of compression applied.

        :Parameters:

            construct:

        :Returns:

            `tuple`

        """
        # For a constructs with data, work out the compression type
        # from the data itself.
        if not self.is_domain(construct):
            data = construct.get_data(None)
            if data is None:
                return ""

            return data.get_compression_type()

        # For a domain construct, work out the compression type from
        # its metadata constucts.
        constructs = self.get_constructs(construct, data=True)

        compression_types = set()

        for c in constructs.values():
            data = c.get_data(None)
            if data is None:
                continue

            #            try:
            #                geometry = c.has_geometry()
            #            except AttributeError:
            #                pass
            #            else:
            #                if geometry:
            #                    # This construct has geometry cells, so does not
            #                    # count as being compressed, for the current
            #                    # purpose.
            #                    continue
            #            # --- End: try

            compression_types.add(data.get_compression_type())

        # The order of the following tests matters
        if "gathered" in compression_types:
            return "gathered"

        elif "ragged indexed contiguous" in compression_types or (
            "ragged indexed" in compression_types
            and "ragged contiguous" in compression_types
        ):
            return "ragged indexed contiguous"

        elif "ragged indexed" in compression_types:
            return "ragged indexed"

        elif "ragged contiguous" in compression_types:
            return "ragged contiguous"

        else:
            return ""

    def get_construct_data_axes(self, field, key):
        """Returns the construct keys of the spanned axes.

        That is, returns the construct keys of the domain axis
        constructs spanned by a metadata construct.

        :Parameters:

            field: field or domain construct

            key: `str`

        :Returns:

            `tuple` or `None`
                The axis identifiers (may be an empty tuple), or
                `None` if there is no data.

        """
        try:
            return field.constructs.data_axes()[key]
        except KeyError:
            return None

    def get_construct_type(self, variable):
        """Return the construct type of a variable.

        .. versionadded:: (cfdm) 1.10.0.1

        :Parameters:

            variable: object
                The object to get a construct type from.

        :Returns:

            `str` or `None`
                The construct type, or `None` if the variable is not a
                construct.

        """
        return getattr(variable, "construct_type", None)

    def get_constructs(self, field, axes=(), data=False):
        """Return constructs that span particular axes.

        If no axes are specified then all constructs are returned.

        If axes are specified then constructs whose data arrays span those
        axes, and possibly other axes, are returned.

        :Parameters:

            field: `Field` or `Domain`

            axes: sequence of `str`

            data: `bool`
                If True then only return constructs that can contain data.

        :Returns:

            `dict`

        """
        if data:
            return field.constructs.filter_by_data(todict=True)

        return field.constructs.filter_by_axis(
            *axes, axis_mode="and", todict=True
        )

    def get_coordinate_reference_coordinates(self, coordinate_reference):
        """Return the coordinates of a coordinate reference construct.

        :Parameters:

            coordinate_reference: coordinate reference construct

        :Returns:

            `set`

        """
        return coordinate_reference.coordinates()

    def get_coordinate_conversion_parameters(self, coordinate_reference):
        """Gets the coordinate reference conversion parameters.

        Specifically, returns the coordinate conversion parameters of a
        coordinate reference construct.

        :Parameters:

            coordinate_reference: coordinate reference construct

        :Returns:

            `dict`

        """
        return coordinate_reference.coordinate_conversion.parameters()

    def get_coordinate_references(self, field):
        """Return all of the coordinate reference constructs of a field.

        :Parameters:

            field: field construct

        :Returns:

        """
        return field.coordinate_references()

    def get_coordinates(self, f):
        """Return all of the coordinate constructs of a field or domain.

        :Parameters:

            f: field or domain construct

        :Returns:

        """
        return f.coordinates()

    def get_data_calendar(self, data, default=None):
        """Return the calendar of date-time data.

        :Parameters:

            data: `data instance

            default: optional

        :Returns:

            `str`

        """
        return data.get_calendar(default=default)

    def get_data_compressed_axes(self, data):
        """Return the indices of the axes that have been compressed.

        :Parameters:

            data: data instance

        :Returns:

            `list` of `int`

        """
        return data.get_compressed_axes()

    def get_data_ndim(self, parent):
        """Return the number of dimensions spanned by the data array.

        :Parameters:

            parent:
                The object containing the data array.

        :Returns:

            `int`
                The number of dimensions spanned by the data array.

        **Examples**

        >>> w = cfdm.implementation()
        >>> d = cfdm.DimensionCoordinate(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data=cfdm.Data(range(180))
        ... )
        >>> d
        <DimensionCoordinate: latitude(180) degrees_north>
        >>> w.get_data_ndim(d)
        1

        >>> b = cfdm.Bounds(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data=cfdm.Data(numpy.arange(360).reshape(180, 2))
        ... )
        >>> b
        <Bounds: latitude(180, 2) degrees_north>
        >>> w.get_data_ndim(b)
        2

        """
        return parent.data.ndim

    def get_data_shape(self, parent, isdata=False):
        """Return the shape of the data array.

        :Parameters:

            parent:
                The object containing the data array.

            isdata: `bool`
                If True then the parent is already a data object

        :Returns:

            `tuple`
                The shape of the data array.

        **Examples**

        >>> w = cfdm.implementation()
        >>> d = cfdm.DimensionCoordinate(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data=cfdm.Data(range(180))
        ... )
        >>> d
        <DimensionCoordinate: latitude(180) degrees_north>
        >>> w.get_data_shape(d)
        (180,)

        >>> b = cfdm.Bounds(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data=cfdm.Data(numpy.arange(360).reshape(180, 2))
        ... )
        >>> b
        <Bounds: latitude(180, 2) degrees_north>
        >>> w.get_data_shape(b)
        (180, 2)

        """
        if isdata:
            return parent.shape

        return parent.data.shape

    def get_data_size(self, parent):
        """Return the number of elements in the data array.

        :Parameters:

            parent:
                The object containing the data array.

        :Returns:

            `int`
                The number of elements in the data array.

        **Examples**

        >>> w = cfdm.implementation()
        >>> d = cfdm.DimensionCoordinate(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data=cfdm.Data(range(180))
        ... )
        >>> d
        <DimensionCoordinate: latitude(180) degrees_north>
        >>> w.get_data_size(d)
        180

        >>> b = cfdm.Bounds(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data=cfdm.Data(numpy.arange(360).reshape(180, 2))
        ... )
        >>> b
        <Bounds: latitude(180, 2) degrees_north>
        >>> w.get_data_size(b)
        360

        """
        return parent.data.size

    def get_data_units(self, data, default=None):
        """Return the units of data.

        :Parameters:

            data: data instance

            default: optional

        :Returns:

            `str`

        """
        return data.get_units(default=default)

    def get_datum(self, coordinate_reference):
        """Return the datum of a coordinate reference construct.

        :Parameters:

              coordinate_reference: coordinate reference construct

        :Returns:

            Datum component

        """
        return coordinate_reference.datum

    def get_datum_parameters(self, ref):
        """Returns coordinate reference datum parameter-valued terms.

        :Parameters:

            coordinate_reference: `CoordinateReference`

        :Returns:

            `dict`

        """
        return ref.datum.parameters()

    def get_dimension_coordinates(self, field):
        """Return all of the dimension coordinate constructs of a field.

        :Parameters:

            field: field construct

        :Returns:

        """
        return field.dimension_coordinates()

    def get_domain_ancillaries(self, field):
        """Return all of the domain ancillary constructs of a field.

        :Parameters:

            field: field construct

        :Returns:

        """
        return field.domain_ancillaries()

    def get_domain_axes(self, field):
        """Return all of the domain axis constructs of a field.

        :Parameters:

            field: field construct

        :Returns:

        """
        return field.domain_axes()

    def get_domain_axis_size(self, field, axis):
        """Return the size a of domrain axis construct.

        :Parameters:

            field: field construct

            axis: `str`

        :Returns:

        """
        return field.domain_axes(todict=True)[axis].get_size()

    def get_sample_dimension_position(self, construct):
        """Returns the position of the compressed data sample dimension.

        :Parameters:

            construct: construct

        :Returns:

            `int`

        """
        return construct.get_data().get_compressed_dimension()

    def nc_get_geometry_variable(self, field, default=None):
        """Return the netCDF variable name of the geometry container.

        .. versionadded:: (cfdm) 1.8.0

        :Parameters:

        :Returns:

            `str`

        """
        return field.nc_get_geometry_variable(default)

    def nc_get_group_attributes(self, field):
        """Returns the netCDF sub-group attribtues for the field.

        .. versionadded:: (cfdm) 1.8.6

        :Parameters:

            field: field construct

        :Returns:

            `dict`

        """
        return field.nc_group_attributes()

    def nc_get_variable_groups(self, field):
        """Return the netCDF groups for the field construct.

        .. versionadded:: (cfdm) 1.8.6

        :Parameters:

            field: field construct

        :Returns:

            `tuple`

        """
        return field.nc_variable_groups()

    def nc_get_hdf5_chunksizes(self, data):
        """Get the HDF5 chunking strategy for the data.

        ..versionadded:: (cfdm) 1.7.2

        :Parameters:

            data: `Data`

        :Returns:

            `tuple` or `int` or `str` or `None`
                The HDF5 chunking strategy.

        """
        return data.nc_hdf5_chunksizes()

    def nc_get_sample_dimension(self, count, default=None):
        """Return the name of the netCDF sample dimension.

        :Returns:

            `str`
                The name of the netCDF sample dimension.

        """
        return count.nc_get_sample_dimension(default=default)

    def nc_is_unlimited_axis(self, field, axis):
        """Whether a domain axis matches a netCDF unlimited dimension.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            field: `Field`

        :Returns:

            `set`
                The selection of domain axis construct identifiers that are
                unlimited.

        """
        domain_axis = field.constructs.get(axis)
        if domain_axis is None:
            return False

        return domain_axis.nc_is_unlimited()

    def nc_set_unlimited_axis(self, field, axis):
        """Sets a domain axis to match a netCDF unlimited dimension.

        .. versionadded:: (cfdm) 1.7.4

        :Parameters:

            field: `Field`

            axis: `str`
                Domain axis construct key.

        :Returns:

            `None`

        """
        domain_axis = field.constructs.get(axis)
        if domain_axis is None:
            return

        domain_axis.nc_set_unlimited(True)

    def nc_get_global_attributes(self, field):
        """Return the netCDF global attributes.

        :Parameters:

            field: field construct

        :Returns:

            `dict`

        """
        return field.nc_global_attributes()

    def nc_set_global_attributes(self, field, attributes):
        """Set netCDF global attributes.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            field: field construct

            attributes: `dict`

        :Returns:

            `None`

        """
        for attr, value in attributes.items():
            field.nc_set_global_attribute(attr, value)

    def nc_set_group_attributes(self, field, attributes):
        """Set netCDF group attributes.

        .. versionadded:: (cfdm) 1.8.6

        :Parameters:

            field: field construct

            attributes: `dict`

        :Returns:

            `None`

        """
        for attr, value in attributes.items():
            field.nc_set_group_attribute(attr, value)

    def nc_set_hdf5_chunksizes(self, data, chunksizes):
        """Set the HDF5 chunking strategy for the data.

        ..versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            data: `Data`

            chunksizes: `int` or `str` or `None` or `dict` or a sequence
                Set the chunking strategy when writing to a netCDF4
                file.

        :Returns:

            `None`

        """
        return data.nc_set_hdf5_chunksizes(chunksizes)

    def equal_components(self, construct0, construct1, ignore_type=False):
        """Whether or not two field construct components are equal.

        A "component" is either a metadata construct or a metadata
        construct component (such as a bounds component).

        .. versionadded::: (cfdm) 1.8.6.0

        :Parameter:

            construct0: construct

            construct1: construct

            ignore_type: `bool`, optional

        :Retuns:

            `bool`

        """
        return construct0.equals(
            construct1, ignore_type=ignore_type, verbose=0
        )

    def equal_constructs(self, construct0, construct1, ignore_type=False):
        """Whether or not two field construct components are equal.

        A "component" is either a metadata construct or a metadata
        construct component (such as a bounds component).

        .. versionadded::: (cfdm) 1.7.0

        :Parameter:

            construct0: construct

            construct1: construct

            ignore_type: `bool`, optional

        :Retuns:

            `bool`

        """
        raise NotImplementedError(
            "Deprecated at version 1.8.6.0. "
            + "Use 'equal_components' instead."
        )

    def equal_properties(self, property_value0, property_value1):
        """Whether or not two property values are equal.

        :Parameter:

            property_value0:

            property_value1:

        :Retuns:

            `bool`

        """
        field = self.get_class("Field")()
        return field._equals(property_value0, property_value1)

    def equal_datums(self, coordinate_reference0, coordinate_reference1):
        """Whether or not two coordinate reference datums are equal.

        :Parameters:

            coordinate_reference0: coordinate reference construct

            coordinate_reference1: coordinate reference construct

        :Returns:

            `bool`

        """
        datum0 = coordinate_reference0.datum
        datum1 = coordinate_reference1.datum
        return datum0.equals(datum1)

    def get_construct_data_size(self, construct):
        """Return the size of a construct's data.

        :Parameters:

            construct: construct

        :Returns:

            `int`

        """
        return construct.data.size

    def nc_get_external(self, parent):
        """Return whether a construct is external.

        :Parameters:

            parent:
                The object

        :Returns:

            `bool`
                Whether the construct is external.

        """
        if not hasattr(parent, "nc_get_external"):
            return False

        return parent.nc_get_external()

    def get_field_ancillaries(self, field):
        """Return the field ancillaries of a field.

        :Parameters:

            field:
                The field object.

        :Returns:

            `dict`
                A dictionary whose values are field ancillary objects, keyed
                by unique identifiers.

        **Examples**

        >>> w = cfdm.implementation()
        >>> f = cfdm.example_field(1)
        >>> w.get_field_ancillaries(f)
        <Constructs: field_ancillary(1)>

        """
        return field.field_ancillaries()

    def get_field_data_axes(self, field):
        """Returns the construct keys of the field's data dimensions.

        :Parameters:

            field: field construct

        :Returns:

            `tuple`

        """
        return field.get_data_axes()

    def get_filenames(self, parent, normalise=True):
        """Return the name of the file or files containing the data.

        :Parameters:

            parent:

            normalise: `bool`, optional
                If True (the default) then normalise the filenames by
                applying any text substitutions and resolving the name
                to an absolute path. If False then neither of these is
                carried out.

                .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `set`

        """
        return parent.get_filenames(normalise=normalise)

    def get_data_max(self, parent):
        """Use `get_data_maximum` instead (since cfdm version 1.8.0)."""
        raise NotImplementedError(
            "Use `get_data_maximum` instead (since cfdm version 1.8.0)."
        )

    def get_data_maximum(self, parent):
        """Return the maximum value of the data.

        :Parameters:

            parent:

        :Returns:

            Scalar `Data` instance

        """
        return parent.data.max(squeeze=True)

    def get_data_sum(self, parent):
        """Return the sum of the data.

        :Parameters:

            parent:

        :Returns:

            Scalar `Data` instance

        """
        return parent.data.sum(squeeze=True)

    def get_count(self, construct):
        """Return the count variable of compressed data.

        :Parameters:

            construct: construct

        :Returns:

            Count variable or `None`

        """
        if not self.is_domain(construct):
            return construct.get_data().get_count(default=None)

        # For a domain construct, get the count variable from its
        # metadata constucts.
        return self._get_domain_compression_variable("count", construct)

    def get_index(self, construct):
        """Return the index variable of compressed data.

        :Parameters:

            construct: construct

        :Returns:

            Index variable or `None`

        """
        if not self.is_domain(construct):
            return construct.get_data().get_index(default=None)

        # For a domain construct, get the index variable from its
        # metadata constucts.
        return self._get_domain_compression_variable("index", construct)

    def get_inherited_properties(self, parent):
        """Return all inherited properties.

        .. versionadded:: (cfdm) 1.8.6

        :Parameters:

            parent:
                The object that inherits the properties.

        :Returns:

            `dict`
                The inherited property names and their values

        """
        return parent.inherited_properties()

    def get_interior_ring(self, construct):
        """Return the interior ring variable of geometry coordinates.

        :Parameters:

            construct: construct

        :Returns:

            Interior ring variable or `None`

        """
        return construct.get_interior_ring(default=None)

    def get_interpolation_name(self, construct, default=None):
        """Return the interior ring variable of geometry coordinates.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            construct: construct

            default:

        :Returns:

            `str`

        """
        try:
            return self.get_data_source(construct).get_interpolation_name(
                default
            )
        except AttributeError:
            return default

    def get_list(self, construct):
        """Return the list variable of compressed data.

        :Parameters:

            construct: construct

        :Returns:

            List variable or `None`

        """
        if not self.is_domain(construct):
            return construct.get_data().get_list(default=None)

        # For a domain construct, get the list variable from its
        # metadata constucts.
        return self._get_domain_compression_variable("list", construct)

    def get_measure(self, cell_measure):
        """Return the measure property of a cell measure construct.

        :Parameters:

            cell_measure:
                The cell measure object.

        :Returns:

            `str` or `None`
                The measure property, or `None` if it has not been set.

        **Examples**

        >>> w = cfdm.implementation()
        >>> c = cfdm.CellMeasure(
        ...     measure='area', properties={'units': 'km2'},
        ...     data=cfdm.Data(numpy.arange(73*96).reshape(73, 96))
        ... )
        >>> c
        <CellMeasure: measure:area(73, 96) km2>
        >>> w.get_measure(c)
        'area'

        """
        return cell_measure.get_measure(default=None)

    def get_original_filenames(self, parent):
        """Get the original names of the files containing the construct.

        :Parameters:

            parent: Construct
                The construct to query of original filenames.

        :Returns:

            `set`
                The original filenames

        """
        return parent.get_original_filenames()

    def nc_get_dimension(self, parent, default=None):
        """Return the netCDF variable name.

        :Parameters:

            parent:
                The object containing the data array.

            default: `str`, optional

        :Returns:

            `str`
                The netCDF dimension name.

        """
        return parent.nc_get_dimension(default=default)

    def nc_get_variable(self, construct, default=None):
        """Return the netCDF variable name.

        :Parameters:

            parent:

            default: `str`, optional

        :Returns:

            `str`
                The netCDF variable name.

        """
        return construct.nc_get_variable(default=default)

    def get_node_count(self, construct):
        """Return the node count variable of geometry coordinates.

        .. versionadded:: (cfdm) 1.8.0

        :Parameters:

            construct: construct

        :Returns:

            Node count variable or `None`

        """
        return construct.get_node_count(default=None)

    def get_part_node_count(self, construct):
        """Return the part node count variable of geometry coordinates.

        .. versionadded:: (cfdm) 1.8.0

        :Parameters:

            construct: construct

        :Returns:

            Part node count variable or `None`

        """
        return construct.get_part_node_count(default=None)

    def get_properties(self, parent):
        """Return all properties.

        :Parameters:

            parent:
                The object containing the properties.

        :Returns:

            `dict`
                The property names and their values

        **Examples**

        >>> w = cfdm.implementation()
        >>> d = cfdm.DimensionCoordinate(
        ...     properties={
        ...         'standard_name': 'latitude',
        ...         'units': 'degrees_north',
        ...         'foo': 'bar'
        ...     },
        ...     data=cfdm.Data(range(180))
        ... )
        >>> d
        <DimensionCoordinate: latitude(180) degrees_north>
        >>> w.get_properties(d)
        {'standard_name': 'latitude', 'units': 'degrees_north', 'foo': 'bar'}

        """
        try:
            return parent.properties()
        except AttributeError:
            return {}

    def get_property(self, construct, prop, default=None):
        """Return a property of a construct.

        :Parameters:

            construct:

            prop: `str`

            default: optional

        :Returns:

                The property value.

        """
        try:
            return construct.get_property(prop, default=default)
        except AttributeError:
            return default

    def get_geometry(self, construct, default=None):
        """Return the geometry type of coordinates.

        :Parameters:

            construct:

            default: optional

        :Returns:

            `str` or `None`
                The geometry type.

        """
        return construct.get_geometry(default=default)

    def get_data(self, parent, default=None):
        """Return the data array.

        :Parameters:

            parent:
                The object containing the data array.

        :Returns:

                The data.

        **Examples**

        >>> w = cfdm.implementation()
        >>> d = cfdm.DimensionCoordinate(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data=cfdm.Data(range(180))
        ... )
        >>> d
        <DimensionCoordinate: latitude(180) degrees_north>
        >>> w.get_data(d)
        <Data(180): [0, ..., 179] degrees_north>

        >>> b = cfdm.Bounds(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data=cfdm.Data(numpy.arange(360).reshape(180, 2))
        ... )
        >>> b
        <Bounds: latitude(180, 2) degrees_north>
        >>> w.get_data(b)
        <Data(180, 2): [[0, ..., 359]] degrees_north>

        """
        try:
            return parent.get_data(default=default)
        except AttributeError:
            return default

    def get_data_axes(self, parent, key, default=None):
        """Get domain axis identifiers.

        :Parameters:

            parent: `Field` or `Domain`

            key: `str` or `None`
                The identifier of the constuct, or `None` to get
                `Field` axes.

        :Returns:

            `tuple` of `str`
                The domain axis identifiers.

        """
        if key is None:
            return parent.get_data_axes(default=default)

        return parent.get_data_axes(key, default=default)

    def get_data_source(self, parent, default=None):
        """Return the data array.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            parent:
                The object containing the data array.

        :Returns:

                The data.

        """
        data = self.get_data(parent, default=None)
        if data is None:
            return default

        return data.source(default)

    def get_tie_points(self, construct, default=None):
        """TODO.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            parent:
                The object containing the data array.

        :Returns:

                The data.

        """
        data = self.get_data_source(construct, None)
        if data is None:
            return default

        return data.source(default)

    def initialise_AuxiliaryCoordinate(self):
        """Return an auxiliary coordinate construct.

        :Returns:

            Auxiliary coordinate construct

        """
        cls = self.get_class("AuxiliaryCoordinate")
        return cls()

    def initialise_Bounds(self):
        """Return a bounds component.

        :Returns:

            Bounds component

        """
        cls = self.get_class("Bounds")
        return cls()

    def initialise_CellConnectivityArray(self, **kwargs):
        """Return a cell connectivity array.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            kwargs: optional
                Parameters for intialising the node connectivity
                array, which are passed to
                `CellConnectivityArray.__init__`.

        :Returns:

            `CellConnectivityArray`

        """
        cls = self.get_class("CellConnectivityArray")
        return cls(**kwargs)

    def initialise_CellMeasure(self, measure=None):
        """Return a cell measure construct.

        :Parameters:

            measure: `str`, optional

        :Returns:

            Cell measure construct

        """
        cls = self.get_class("CellMeasure")
        return cls(measure=measure)

    def initialise_CellMethod(self, axes=None, method=None, qualifiers=None):
        """Return a cell method construct.

        :Parameters:

            axes: optional

            method: `str`, optional

            qualifiers: `dict`, optional

        :Returns:

            Cell method construct

        """
        cls = self.get_class("CellMethod")
        return cls(axes=axes, method=method, qualifiers=qualifiers)

    def initialise_AggregatedArray(self, **kwargs):
        """Return a `AggregatedArray` instance.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            kwargs: optional
                Initialisation parameters to pass to the new instance.

        :Returns:

            `AggregatedArray`

        """
        cls = self.get_class("AggregatedArray")
        return cls(**kwargs)

    def initialise_CoordinateConversion(
        self, domain_ancillaries=None, parameters=None
    ):
        """Return a coordinate conversion component.

        :Parameters:

            domain_ancillaries: optional

            parameters: optioanl

        :Returns:

            Coordinate conversion component

        """
        cls = self.get_class("CoordinateConversion")
        return cls(
            domain_ancillaries=domain_ancillaries, parameters=parameters
        )

    def initialise_CoordinateReference(self):
        """Return a coordinate reference construct.

        :Returns:

            Coordinate reference construct

        """
        cls = self.get_class("CoordinateReference")
        return cls()

    def initialise_Count(self):
        """Return a count variable.

        :Returns:

            Count variable

        """
        cls = self.get_class("Count")
        return cls()

    def initialise_Data(
        self,
        array=None,
        units=None,
        calendar=None,
        copy=True,
        **kwargs,
    ):
        """Return a data instance.

        :Parameters:

            array: optional

            units: optional

            calendar: optional

            copy: `bool`, optional

            kwargs: optional

        :Returns:

            Data instance

        """
        cls = self.get_class("Data")
        return cls(
            array=array,
            units=units,
            calendar=calendar,
            copy=copy,
            **kwargs,
        )

    def initialise_Datum(self, parameters=None):
        """Return a coordinate conversion component.

        :Parameters:

            parameters: optioanl

        :Returns:

            Datum component

        """
        cls = self.get_class("Datum")
        return cls(parameters=parameters)

    def initialise_DimensionCoordinate(
        self,
        properties=None,
        data=None,
        bounds=None,
        interior_ring=None,
        copy=True,
    ):
        """Return a dimension coordinate construct.

        :Parameters:

            properties: `dict`, optional

            data: optional

            bounds: optional

            interior_ring: optional

            copy: `bool`, optional

        :Returns:

            Dimension coordinate construct

        """
        cls = self.get_class("DimensionCoordinate")
        return cls(
            properties=properties,
            data=data,
            bounds=bounds,
            interior_ring=interior_ring,
            copy=copy,
        )

    def initialise_DimensionCoordinate_from_AuxiliaryCoordinate(
        self, auxiliary_coordinate=None, copy=True
    ):
        """Returns a dimension coordinate from an auxiliary coordinate.

        Specifically, returns a dimension coordinate construct
        insitialised from an auxiliary coordinate construct.

        :Parameters:

            auxiliary_coordinate: auxiliary coordinate construct

            copy: `bool`,optional

        :Returns:

            Dimension coordinate construct

        """
        cls = self.get_class("DimensionCoordinate")
        return cls(source=auxiliary_coordinate, copy=copy)

    def initialise_Domain(self):
        """Return a domain construct.

        :Returns:

            Domain construct

        """
        cls = self.get_class("Domain")
        return cls()

    def initialise_DomainAncillary(self):
        """Return a domain ancillary construct.

        :Returns:

            Domain ancillary construct

        """
        cls = self.get_class("DomainAncillary")
        return cls()

    def initialise_DomainAxis(self, size=None):
        """Return a domain ancillary construct.

        :Returns:

            Domain ancillary construct

        """
        cls = self.get_class("DomainAxis")
        return cls(size=size)

    def initialise_DomainTopology(self, **kwargs):
        """Return a domain topology construct.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            kwargs: optional
                Parameters for intialising the domain topology
                construct, which are passed to
                `DomainTopology.__init__`.

        :Returns:

            `DomainTopology`
                The domain topology construct.

        """
        cls = self.get_class("DomainTopology")
        return cls(**kwargs)

    def initialise_CellConnectivity(self, **kwargs):
        """Return a cell connectivity construct.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            kwargs: optional
                Parameters for intialising the cell connectivity
                construct, which are passed to
                `CellConnectivity.__init__`.

        :Returns:

            `CellConnectivity`
                The cell connectivity construct.

        """
        cls = self.get_class("CellConnectivity")
        return cls(**kwargs)

    def initialise_Field(self):
        """Return a field qconstruct.

        :Returns:

            Field construct

        """
        cls = self.get_class("Field")
        return cls()

    def initialise_FieldAncillary(self):
        """Return a field ancillary construct.

        :Returns:

            Field ancillary construct

        """
        cls = self.get_class("FieldAncillary")
        return cls()

    def initialise_GatheredArray(
        self,
        compressed_array=None,
        ndim=None,
        shape=None,
        size=None,
        compressed_dimension=None,
        compressed_dimensions=None,
        list_variable=None,
    ):
        """Return a gathered array instance.

        :Parameters:

            compressed_array: optional

            ndim: `int`, optional

            shape: sequence of `int`, optional

            size: `int, optional

            compressed_dimensions: sequence of `int`, optional
                The position of the compressed dimension in the
                compressed array.

                 .. versionadded:: (cfdm) 1.10.0.0

            list_variable: optional

            compressed_dimension: deprecated at version 1.10.0.0
                Use the *compressed_dimensions* parameter instead.

        :Returns:

            Gathered array

        """
        cls = self.get_class("GatheredArray")
        return cls(
            compressed_array=compressed_array,
            ndim=ndim,
            shape=shape,
            size=size,
            compressed_dimensions=compressed_dimensions,
            list_variable=list_variable,
        )

    def initialise_SubsampledArray(
        self,
        interpolation_name=None,
        compressed_array=None,
        shape=None,
        tie_point_indices=None,
        computational_precision=None,
        interpolation_description=None,
        parameters=None,
        parameter_dimensions=None,
        dependent_tie_points=None,
        dependent_tie_point_dimensions=None,
    ):
        """Return a subsampled array instance.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `set_dependent_tie_points`,
                     `set_dependent_tie_point_dimensions`

        :Parameters:

            interpolation_name: `str`

            interpolation_description: `str`, optional

            compressed_array: optional

            shape: sequence of `int`, optional

            computational_precision: `str`, optional
                The floating-point arithmetic precision used during
                the preparation and validation of the compressed
                coordinates.

            tie_point_indices: `dict`, optional

            parameters: `dict`, optional

            parameter_dimensions: `dict`, optional

            parameters: `dict`, optional

            parameter_dimensions: `dict`, optional

            dependent_tie_points: `dict`, optional

            dependent_tie_point_dimensions: `dict`, optional

        :Returns:

            Subsampled array

        """
        return self.get_class("SubsampledArray")(
            interpolation_name=interpolation_name,
            compressed_array=compressed_array,
            shape=shape,
            tie_point_indices=tie_point_indices,
            computational_precision=computational_precision,
            interpolation_description=interpolation_description,
            parameters=parameters,
            parameter_dimensions=parameter_dimensions,
            dependent_tie_points=dependent_tie_points,
            dependent_tie_point_dimensions=dependent_tie_point_dimensions,
        )

    def initialise_Index(self):
        """Return an index variable.

        :Returns:

            Index variable

        """
        cls = self.get_class("Index")
        return cls()

    def initialise_InteriorRing(self):
        """Return an interior ring variable.

        :Returns:

            Interior ring variable

        """
        cls = self.get_class("InteriorRing")
        return cls()

    def initialise_InterpolationParameter(self):
        """Return an interpolation parameter variable.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            Interpolation parameter variable

        """
        cls = self.get_class("InterpolationParameter")
        return cls()

    def initialise_List(self):
        """Return a list variable.

        :Returns:

            List variable

        """
        cls = self.get_class("List")
        return cls()

    def initialise_PointTopologyArray(self, **kwargs):
        """Return a point topology array.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            kwargs: optional
                Parameters for intialising the point topology array.
                which are passed to `PointTopologyArray.__init__`.

        :Returns:

            `PointTopologyArray`

        """
        cls = self.get_class("PointTopologyArray")
        return cls(**kwargs)

    def initialise_TiePointIndex(self):
        """Return an index variable.

        :Returns:

            Index variable

        """
        cls = self.get_class("TiePointIndex")
        return cls()

    def initialise_NetCDF4Array(self, **kwargs):
        """Return a `NetCDF4Array` instance.

        :Parameters:

            kwargs: optional
                Initialisation parameters to pass to the new instance.

                .. versionadded:: (cfdm) 1.11.2.0

        :Returns:

            `NetCDF4Array`

        """
        cls = self.get_class("NetCDF4Array")
        return cls(**kwargs)

    def initialise_H5netcdfArray(self, **kwargs):
        """Return a `H5netcdfArray` instance.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            kwargs: optional
                Initialisation parameters to pass to the new instance.

        :Returns:

            `H5netcdfArray`

        """
        cls = self.get_class("H5netcdfArray")
        return cls(**kwargs)

    def initialise_BoundsFromNodesArray(self, **kwargs):
        """Return a node bounds array.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            kwargs: optional
                Parameters for intialising the node bounds array.
                which are passed to `BoundsFromNodesArray.__init__`.

        :Returns:

            `BoundsFromNodesArray`

        """
        cls = self.get_class("BoundsFromNodesArray")
        return cls(**kwargs)

    def initialise_NodeCountProperties(self):
        """Return a node count properties variable.

        :Returns:

            Node count properties bariable

        """
        cls = self.get_class("NodeCountProperties")
        return cls()

    def initialise_PartNodeCount(self):
        """Return a part node count properties variable.

        :Returns:

            Part node count properties variable

        """
        cls = self.get_class("PartNodeCountProperties")
        return cls()

    def initialise_RaggedContiguousArray(
        self,
        compressed_array=None,
        ndim=None,
        shape=None,
        size=None,
        count_variable=None,
    ):
        """Return a ragged contigous array instance.

        :Parameters:

            compressed_array: optional

            ndim: `int`, optional
                Deprecated at version 1.10.0.0. Ignored if set.

            shape: sequence of `int`, optional

            size: `int, optional
                Deprecated at version 1.10.0.0. Ignored if set.

            count_variable: optional

        :Returns:

            Ragged contigous array

        """
        cls = self.get_class("RaggedContiguousArray")
        return cls(
            compressed_array=compressed_array,
            shape=shape,
            count_variable=count_variable,
        )

    def initialise_RaggedIndexedArray(
        self,
        compressed_array=None,
        ndim=None,
        shape=None,
        size=None,
        index_variable=None,
    ):
        """Return a ragged indexed array instance.

        :Parameters:

            compressed_array: optional

            ndim: `int`, optional
                Deprecated at version 1.10.0.0. Ignored if set.

            shape: sequence of `int`, optional

            size: `int, optional
                Deprecated at version 1.10.0.0. Ignored if set.

            index_variable: optional

        :Returns:

            Ragged indexed array

        """
        cls = self.get_class("RaggedIndexedArray")
        return cls(
            compressed_array=compressed_array,
            #            ndim=ndim,
            shape=shape,
            #            size=size,
            index_variable=index_variable,
        )

    def initialise_RaggedIndexedContiguousArray(
        self,
        compressed_array=None,
        ndim=None,
        shape=None,
        size=None,
        count_variable=None,
        index_variable=None,
    ):
        """Return a ragged indexed contiguous array instance.

        :Parameters:

            compressed_array: optional

            ndim: `int`, optional
                Deprecated at version 1.10.0.0. Ignored if set.

            shape: sequence of `int`, optional

            size: `int, optional
                Deprecated at version 1.10.0.0. Ignored if set.

            count_variable: optional

            index_variable: optional

        :Returns:

             Ragged indexed contiguous array

        """
        cls = self.get_class("RaggedIndexedContiguousArray")
        return cls(
            compressed_array=compressed_array,
            shape=shape,
            count_variable=count_variable,
            index_variable=index_variable,
        )

    def is_climatology(self, coordinate):
        """Whether or not the coordinate represent climatologies.

        :Parameters:

            coordinate: coordinate construct

        :Returns:

            `bool`

        """
        return bool(coordinate.get_geometry(None) == "climatology")

    def is_domain(self, construct):
        """Return True if the construct is a domain construct.

        :Parameters:

            construct: Construct

        :Returns:

            `bool`

        """
        return getattr(construct, "construct_type", None) == "domain"

    def is_field(self, construct):
        """Return True if the construct is a field construct.

        :Parameters:

            construct: Construct

        :Returns:

            `bool`

        """
        return getattr(construct, "construct_type", None) == "field"

    def is_geometry(self, coordinate):
        """Return True if the coordinate bounds are geometries.

        ..versionadded:: (cfdm) 1.8.0

        :Parameters:

            coordinate:
                The coordinate construct.

        :Returns:

            `bool`
                 True if the coordinate bounds are geometries, otherwise
                 False.

        """
        return bool(
            coordinate.get_geometry(None) in ("point", "line", "polygon")
        )

    def is_masked(self, data):
        """Whether or not the data has any masked values.

        .. versionadded:: (cfdm) 1.8.0

        :Parameters:

            data: `Data`

        :Returns:

            `bool`
                Whether or not the data has any masked values.

        """
        return data.mask.any()

    def nc_set_instance_dimension(self, variable, ncdim):
        """Set the netCDF instance dimension name.

        :Parameters:

            variable:

            ncdim: `str` or `None`
                The netCDF dimension name. If `None` then the name is
                not set.

        :Returns:

            `None`

        """
        if ncdim is not None:
            variable.nc_set_instance_dimension(ncdim)

    def nc_set_sample_dimension(self, variable, ncdim):
        """Set the netCDF sample dimension name.

        :Parameters:

            variable:

            ncdim: `str` or `None`
                The netCDF dimension name. If `None` then the name is not
                set.

        :Returns:

            `None`

        """
        if ncdim is not None:
            variable.nc_set_sample_dimension(ncdim)

    def nc_set_interpolation_subarea_dimension(self, variable, ncdim):
        """Set the netCDF sample dimension name.

        :Parameters:

            variable:

            ncdim: `str` or `None`
                The netCDF interpolation subarea dimension name. If
                `None` then the name is not set.

        :Returns:

            `None`

        """
        if ncdim is not None:
            variable.nc_set_interpolation_subarea_dimension(ncdim)

    def nc_set_subsampled_dimension(self, variable, ncdim):
        """Set the netCDF subsampled dimension name.

        :Parameters:

            variable:

            ncdim: `str` or `None`
                The netCDF dimension name. If `None` then the name is
                not set.

        :Returns:

            `None`

        """
        if ncdim is not None:
            variable.nc_set_subsampled_dimension(ncdim)

    def set_auxiliary_coordinate(
        self, parent, construct, axes, copy=True, **kwargs
    ):
        """Insert a auxiliary coordinate object into a field.

        :Parameters:

            parent: `Field` or `Domain`
               On what to set the construct

            construct: auxiliary coordinate construct

            axes: `tuple`

            copy: `bool`, optional

            kwargs: optional
                Additional parameters to `Field.set_construct` that
                may be used by subclasses.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `str`

        """
        return self.set_construct(
            parent, construct, axes=axes, copy=copy, **kwargs
        )

    def set_bounds(self, construct, bounds, copy=True):
        """Set the bounds component of a construct.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            construct: construct

            bounds: bounds component

            copy: `bool`, optional

        :Returns:

            `str`
                Return an empty string if the bounds were set
                successfully, otherwise return a non-empty string
                describing how the setting of the bounds failed.

        """
        try:
            construct.set_bounds(bounds, copy=copy)
        except Exception as error:
            if not error:
                error = f"Could not set {bounds!r} on {construct!r}"

            return error

        return ""

    def set_cell_measure(self, parent, construct, axes, copy=True, **kwargs):
        """Insert a cell_measure object into a field.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            parent: `Field` or `Domain`
               On what to set the construct

            construct: cell measure construct

            axes: `tuple`

            copy: `bool`, optional

            kwargs: optional
                Additional parameters to the `set_construct` of
                *parent* that may be used by subclasses.

                .. versionadded:: (cfdm) 1.10.0.2

        :Returns:

            `str`

        """
        return self.set_construct(
            parent, construct, axes=axes, copy=copy, **kwargs
        )

    def set_cell_method(self, field, construct, copy=True, **kwargs):
        """Insert a cell_method object into a field.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            field: `Field`
               On what to set the construct

            construct: cell method construct

            copy: `bool`, optional

            kwargs: optional
                Additional parameters to the `set_construct` of
                *field* that may be used by subclasses.

                .. versionadded:: (cfdm) 1.10.0.2

        :Returns:

            `str`

        """
        return self.set_construct(field, construct, copy=copy, **kwargs)

    def set_cell_method_axes(self, cell_method, axes):
        """Set the axes of a cell method construct.

        :Parameters:

            construct: cell method construct

            axes:

        :Returns:

            `None`

        """
        cell_method.set_axes(axes)

    def set_cell_method_method(self, cell_method, method):
        """Set the method of a cell method construct.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            cell_method: cell method construct

            method: `str`

        :Returns:

            `None`

        """
        cell_method.set_method(method)

    def set_cell_connectivity(
        self, parent, construct, axes, copy=True, **kwargs
    ):
        """Insert a cell connectivity object into a field.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            parent: `Field` or `Domain`
               On what to set the construct

            construct: `CellConnectivity`

            axes: `tuple`

            copy: `bool`, optional

            kwargs: optional
                Additional parameters to the `set_construct` of
                *parent* that may be used by subclasses.


        :Returns:

            `str`

        """
        return self.set_construct(
            parent, construct, axes=axes, copy=copy, **kwargs
        )

    def set_climatology(self, construct):
        """Set the construct as a climatology.

        .. versionadded:: (cfdm) 1.9.0.0

        :Parameters:

            construct:

        :Returns:

            `None`

        """
        construct.set_climatology(True)

    def set_construct(self, parent, construct, axes=None, copy=True, **kwargs):
        """Insert a construct into a field or domain.

        .. versionadded:: (cfdm) 1.10.0.2

        :Parameters:

            parent: `Field` or `Domain`
               On what to set the construct

            construct:
                The construct to set.

            axes: `tuple` or `None`, optional
                The construct domain axes, if applicable.

            copy: `bool`, optional
                Whether or not to set a copy of *construct*.

            kwargs: optional
                Additional parameters to the `set_construct` of
                *parent* that may be used by subclasses.

        :Returns:

            `str`
                The construct identifier.

        """
        return parent.set_construct(construct, axes=axes, copy=copy, **kwargs)

    def set_coordinate_conversion(
        self, coordinate_reference, coordinate_conversion
    ):
        """Set the coordinate conversion coordinate reference construct.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            coordinate_reference: coordinate reference construct

            coordinate_conversion: coordinate conversion component

        :Returns:

            `None`

        """
        coordinate_reference.set_coordinate_conversion(coordinate_conversion)

    def set_coordinate_reference(self, parent, construct, copy=True, **kwargs):
        """Insert a coordinate reference object into a field.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            parent: `Field` or `Domain`
               On what to set the construct

            construct: coordinate reference construct

            copy: `bool`, optional

            kwargs: optional
                Additional parameters to the `set_construct` of
                *parent* that may be used by subclasses.

                .. versionadded:: (cfdm) 1.10.0.2

        :Returns:

            `str`

        """
        return self.set_construct(parent, construct, copy=copy, **kwargs)

    def set_coordinate_reference_coordinates(
        self, coordinate_reference, coordinates
    ):
        """Set the coordinates of a coordinate reference construct.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            coordinate_reference: coordinate reference construct

            coordinates: sequence of `str`

        :Returns:

            `None`

        """
        coordinate_reference.set_coordinates(coordinates)

    def set_coordinate_reference_coordinate(
        self, coordinate_reference, coordinate
    ):
        """Set a coordinate of a coordinate reference construct.

        :Parameters:

            coordinate_reference: coordinate reference construct

            coordinate: `str`

        :Returns:

            `None`

        """
        coordinate_reference.set_coordinate(coordinate)

    def set_data(self, construct, data, axes=None, copy=True):
        """The the data instance of a construct.

        :Parameters:

            construct: construct

            data: data instance

            axes: `tuple`, optional

            default: optional

        :Returns:

            `None`

        """
        if axes is None:
            construct.set_data(data, copy=copy)
        else:
            construct.set_data(data, axes=axes, copy=copy)

    def set_datum(self, coordinate_reference, datum):
        """Insert a datum object into a coordinate reference construct.

        :Parameters:

            coordinate_referece: coordinate reference construct

            datum: datum component

        :Returns:

            `None`

        """
        coordinate_reference.set_datum(datum)

    def set_dimension_coordinate(
        self, parent, construct, axes, copy=True, **kwargs
    ):
        """Insert a dimension coordinate object into a field.

        :Parameters:

            parent: `Field` or `Domain`
               On what to set the construct

            construct: dimension coordinate construct

            axes: `tuple`

            copy: `bool`, optional

            kwargs: optional
                Additional parameters to the `set_construct` of
                *parent* that may be used by subclasses.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `str`

        """
        return self.set_construct(
            parent, construct, axes=axes, copy=copy, **kwargs
        )

    def set_domain_ancillary(
        self, parent, construct, axes, copy=True, **kwargs
    ):
        """Insert a domain ancillary object into a field.

        :Parameters:

            parent: `Field` or `Domain`
               On what to set the construct

            construct: domain ancillary construct`

            axes: `tuple`

            copy: `bool`, optional

            kwargs: optional
                Additional parameters to the `set_construct` of
                *parent* that may be used by subclasses.

                .. versionadded:: (cfdm) 1.10.0.2

        :Returns:

            `str`

        """
        return self.set_construct(
            parent, construct, axes=axes, copy=copy, **kwargs
        )

    def set_domain_axis(self, parent, construct, copy=True, **kwargs):
        """Insert a domain_axis object into a field.

        :Parameters:

            parent: `Field` or `Domain`
               On what to set the construct

            construct: domain axis construct

            copy: `bool`, optional

            kwargs: optional
                Additional parameters to the `set_construct` of
                *parent* that may be used by subclasses.

                .. versionadded:: (cfdm) 1.10.0.2

        :Returns:

            `str`

        """
        return self.set_construct(parent, construct, copy=copy, **kwargs)

    def set_domain_topology(
        self, parent, construct, axes, copy=True, **kwargs
    ):
        """Insert a domain topology object into a field.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            parent: `Field` or `Domain`
               On what to set the construct

            construct: `DomainTopology`

            axes: `tuple`

            copy: `bool`, optional

            kwargs: optional
                Additional parameters to the `set_construct` of
                *parent* that may be used by subclasses.


        :Returns:

            `str`

        """
        return self.set_construct(
            parent, construct, axes=axes, copy=copy, **kwargs
        )

    def set_dependent_tie_points(self, construct, tie_points, dimensions):
        """Set dependent tie points and their dimensions.

        This applies to arrays compressed by subsampling.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            construct:
                The construct containing the ccompressed data. Will be
                changed in-place.

            tie_points: `dict`
                The dependent tie point arrays, keyed by their
                identities.

            dimensions: `dict`
                The dimensions spanned by the tie point ararys, keyed
                by their identities.

        :Returns:

            `None`

        """
        data = self.get_data(construct)
        source = data.source()

        source.set_dependent_tie_points(tie_points)
        source.set_dependent_tie_point_dimensions(dimensions)

        # Reset the construct's data with the new dependent tie point
        # definitions. Currently cfdm doesn't need to do this, but it
        # doesn't do any harm, and downstream applications (like
        # cf-python) may not be able to correctly instantiate the data
        # when all of its elements have not been set.
        data = construct._Data(
            source,
            units=data.get_units(None),
            calendar=data.get_calendar(None),
        )
        construct.set_data(data)

    def set_mesh_id(self, parent, mesh_id):
        """Set a mesh identifier.

        .. versionadded:: (cfdm)  1.11.0.0

        :Parameters:

            parent: construct
                The construct on which to set the mesh id

            mesh_id:
                The mesh identifier.

        :Returns:

            `None`

        """
        parent.set_mesh_id(mesh_id)

    def nc_set_external(self, construct):
        """Set the external status of a construct.

        :Parameters:

            construct:

        :Returns:

            `None`

        """
        construct.nc_set_external(True)

    def set_field_ancillary(self, field, construct, axes, copy=True, **kwargs):
        """Insert a field ancillary object into a field.

        :Parameters:

            field: `Field`
               On what to set the construct

            construct: field ancillary construct

            axes: `tuple`

            copy: `bool`, optional

            kwargs: optional
                Additional parameters to the `set_construct` of
                *field* that may be used by subclasses.

                .. versionadded:: (cfdm) 1.10.0.2

        :Returns:

            `str`

        """
        return self.set_construct(
            field, construct, axes=axes, copy=copy, **kwargs
        )

    def set_geometry(self, coordinate, value):
        """Set the geometry type of a coordinate construct.

        .. versionadded:: (cfdm) 1.8.0

        :Parameters:

            coordinate: coordinate construct

            value: `str`

        :Returns:

            `None`

        """
        coordinate.set_geometry(value)

    def set_inherited_properties(
        self, parent, inherited_properties, copy=True
    ):
        """Set any inherited properties.

        .. versionadded:: (cfdm) 1.8.6

        :Parameters:

            parent:
                The object that inherits the properties.

            inherited_properties: `dict`

            copy: `bool``

        :Returns:

            `None`

        """
        parent.set_properties(inherited_properties, copy=copy)

    def set_interpolation(self, parent, interpolation, copy=True):
        """Set an interpolation variable.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            parent:

            interpolation: Interpolation properties variable

            copy: `bool`, optional

        :Returns:

            `None`

        """
        parent.set_interpolation(interpolation, copy=copy)

    def set_node_count_properties(self, parent, node_count, copy=True):
        """Set a node count properties variable.

        .. versionadded:: (cfdm) 1.8.0

        :Parameters:

            parent:

            node_count: Node count properties variable

            copy: `bool`, optional

        :Returns:

            `None`

        """
        parent.set_node_count(node_count, copy=copy)

    def set_original_filenames(self, parent, filename):
        """Set the original files.

        Set the original files to the *filename* parameter plus any
        original files stored on the data (if present).

        .. versionadded:: (cfdm) 1.10.0.1

        :Parameters:

            parent: Parent construct on which to set original files

            filename: `str`
                The file that contains the corresponding variable.

        :Returns:

            `None`

        """
        if filename is None:
            filenames = ()
        else:
            filenames = (filename,)

        try:
            data = parent.get_data(None)
        except AttributeError:
            pass
        else:
            if data is not None:
                filenames += tuple(data._original_filenames())

        parent._original_filenames(define=set(filenames))

    def set_part_node_count_properties(
        self, parent, part_node_count, copy=True
    ):
        """Set a part node count properties variable.

        .. versionadded:: (cfdm) 1.8.0

        :Parameters:

            parent:

            part_node_count: part node count properties variable

            copy: `bool`, optional

        :Returns:

            `None`

        """
        parent.set_part_node_count(part_node_count, copy=copy)

    def set_interior_ring(self, parent, interior_ring, copy=True):
        """Insert an interior ring array into a coordinate.

        .. versionadded:: (cfdm) 1.8.0

        :Parameters:

            parent:

            interior_ring: interiot ring variable

            copy: `bool`, optional

        :Returns:

            `None`

        :Returns:

            `None`

        """
        parent.set_interior_ring(interior_ring, copy=copy)

    def set_dataset_compliance(self, field, report):
        """Set the dataset compliance report on a field construct.

        ..versionadded:: (cfdm) 1.7

        :Parameters:

            field: field construct

            report: `dict`

        :Returns:

            `None`

        """
        field._set_dataset_compliance(report)

    def nc_set_dimension(self, construct, ncdim):
        """Set the netCDF dimension name.

        :Parameters:

            construct: construct

            ncdim: `str` or `None`
                The netCDF dimension name. If `None` then the name is
                not set.

        :Returns:

            `None`

        """
        if ncdim is not None:
            construct.nc_set_dimension(ncdim)

    def nc_set_geometry_variable(self, field, ncvar):
        """Set the netCDF geometry container variable name.

        :Parameters:

            field: field construct

            ncvar: `str` or `None`
                The netCDF variable name. If `None` then the name is
                not set.

        :Returns:

            `None`

        """
        if ncvar is not None:
            field.nc_set_geometry_variable(ncvar)

    def nc_set_node_coordinate_variable(self, parent, ncvar):
        """Set the netCDF node coordinate variable name.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            parent: `AuxiliaryCoordinate`
                The auxiliary coordinate construct on which to set the
                netCDF node coordinate variable name.

            ncvar: `str` or `None`
                The netCDF node coordinate variable name. If `None`
                then the name is not set.

        :Returns:

            `None`

        """
        if ncvar is not None:
            parent.nc_set_node_coordinate_variable(ncvar)

    def nc_set_variable(self, parent, ncvar):
        """Set the netCDF variable name.

        :Parameters:

            parent:

            ncvar: `str` or `None`
                The netCDF variable name. If `None` then the name is
                not set.

        :Returns:

            `None`

        """
        if ncvar is not None:
            parent.nc_set_variable(ncvar)

    def nc_get_datum_variable(self, ref):
        """Get the netCDF grid mapping variable name for a datum.

        .. versionadded:: (cfdm) 1.7.5

        :Parameters:

            ref: Coordinate reference construct

        :Returns:

            `str` or `None`

        """
        return ref.nc_get_datum_variable(default=None)

    def nc_set_datum_variable(self, ref, ncvar):
        """Set the netCDF grid mapping variable name for a datum.

        .. versionadded:: (cfdm) 1.7.5

        :Parameters:

            ref: Coordinate reference construct

            ncvar: `str` or `None`
                The netCDF variable name. If `None` then the name is not
                set.

        :Returns:

            `None`

        """
        if ncvar is not None:
            ref.nc_set_datum_variable(ncvar)

    def set_properties(self, construct, properties, copy=True):
        """Set construct proporties.

        :Parameters:

            construct:

            properties: `dict`

            copy: `bool`

        :Returns:

            `None`

        """
        construct.set_properties(properties, copy=copy)

    def has_bounds(self, construct):
        """Whether or not a construct has bounds.

        :Parameters:

            construct:

        :Returns:

            `bool`

        """
        return construct.has_bounds()

    def has_datum(self, coordinate_reference):
        """Return True if a coordinate reference has a datum.

        :Parameters:

            coordinate_reference: coordinate reference construct

        :Returns:

            `bool`

        **Examples**

        >>> w = cfdm.implementation()

        >>> c = cfdm.CoordinateReference()
        >>> w.has_datum(c)
        False

        >>> r = cfdm.CoordinateReference(datum=cfdm.Data(1))
        >>> w.has_datum(r)
        True

        """
        return bool(coordinate_reference.datum)

    def has_domain_topology(self, parent):
        """Whether or not a field or domain has a domain topology.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            parent: `Field` or `Domain`
                A field or domain construct

        :Returns:

            `bool`
                Whether or not *parent* has a domain topology.

        """
        try:
            return bool(parent.domain_topologies(todict=True))
        except AttributeError:
            return False

    def has_property(self, parent, prop):
        """Return True if a property exists.

        :Parameters:

            parent:
                The object containing the property.

        :Returns:

            `bool`
                `True` if the property exists, otherwise `False`.

        **Examples**

        >>> w = cfdm.implementation()
        >>> d = cfdm.DimensionCoordinate(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data=cfdm.Data(range(180))
        ... )
        >>> d
        <DimensionCoordinate: latitude(180) degrees_north>
        >>> w.has_property(d, 'units')
        True

        >>> b = cfdm.Bounds(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data=cfdm.Data(numpy.arange(360).reshape(180, 2))
        ... )
        >>> b
        <Bounds: latitude(180, 2) degrees_north>
        >>> w.has_property(b, 'long_name')
        False

        """
        return parent.has_property(prop)

    def squeeze(self, construct, axes=None, inplace=False):
        """Remove size 1 axes from construct data.

        :Parameters:

            construct:
                The construct.

            axes: optional
                The axes to squeeze. If `None` then all size 1 axes
                are removed from the data.

            inplace: `bool`, optional
                If True then do the operation in-place and return
                `None`.

                .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

                The construct with removed axes, or `None` if the
                operation was in-place.

        """
        return construct.squeeze(axes=axes, inplace=inplace)

    def unsqueeze(self, field, inplace=False):
        """Insert size 1 axes into the field data array.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            field: `Field`
                The field construct.

            inplace: `bool`, optional
                If True then do the operation in-place and return
                `None`.

        :Returns:

            `Field` or `None`
                The field with inserted axes, or `None` if the
                operation was in-place.

        """
        return field.unsqueeze(inplace=inplace)


_implementation = CFDMImplementation(
    cf_version=CF(),
    AggregatedArray=AggregatedArray,
    AuxiliaryCoordinate=AuxiliaryCoordinate,
    CellConnectivity=CellConnectivity,
    CellConnectivityArray=CellConnectivityArray,
    CellMeasure=CellMeasure,
    CellMethod=CellMethod,
    CoordinateReference=CoordinateReference,
    DimensionCoordinate=DimensionCoordinate,
    Domain=Domain,
    DomainAncillary=DomainAncillary,
    DomainAxis=DomainAxis,
    DomainTopology=DomainTopology,
    Field=Field,
    FieldAncillary=FieldAncillary,
    Bounds=Bounds,
    InteriorRing=InteriorRing,
    InterpolationParameter=InterpolationParameter,
    CoordinateConversion=CoordinateConversion,
    Datum=Datum,
    List=List,
    Index=Index,
    Count=Count,
    NodeCountProperties=NodeCountProperties,
    PartNodeCountProperties=PartNodeCountProperties,
    Data=Data,
    BoundsFromNodesArray=BoundsFromNodesArray,
    GatheredArray=GatheredArray,
    H5netcdfArray=H5netcdfArray,
    NetCDF4Array=NetCDF4Array,
    PointTopologyArray=PointTopologyArray,
    RaggedContiguousArray=RaggedContiguousArray,
    RaggedIndexedArray=RaggedIndexedArray,
    RaggedIndexedContiguousArray=RaggedIndexedContiguousArray,
    SubsampledArray=SubsampledArray,
    TiePointIndex=TiePointIndex,
)


def implementation():
    """Return a container for the CF data model implementation.

    .. versionadded:: (cfdm) 1.7.0

    .. seealso:: `cfdm.example_field`, `cfdm.read`, `cfdm.write`

    :Returns:

        `CFDMImplementation`
            A container for the CF data model implementation.

    **Examples**

    >>> i = cfdm.implementation()
    >>> i
    <CFDMImplementation: >
    >>> i.classes()
    {'AuxiliaryCoordinate': <class 'cfdm.auxiliarycoordinate.AuxiliaryCoordinate'>,
     'CellMeasure': <class 'cfdm.cellmeasure.CellMeasure'>,
     'CellMethod': <class 'cfdm.cellmethod.CellMethod'>,
     'CoordinateReference': <class 'cfdm.coordinatereference.CoordinateReference'>,
     'DimensionCoordinate': <class 'cfdm.dimensioncoordinate.DimensionCoordinate'>,
     'DomainAncillary': <class 'cfdm.domainancillary.DomainAncillary'>,
     'DomainAxis': <class 'cfdm.domainaxis.DomainAxis'>,
     'Field': <class 'cfdm.field.Field'>,
     'FieldAncillary': <class 'cfdm.fieldancillary.FieldAncillary'>,
     'Bounds': <class 'cfdm.bounds.Bounds'>,
     'InteriorRing': <class 'cfdm.interiorring.InteriorRing'>,
     'CoordinateConversion': <class 'cfdm.coordinateconversion.CoordinateConversion'>,
     'Datum': <class 'cfdm.datum.Datum'>,
     'Data': <class 'cfdm.data.data.Data'>,
     'GatheredArray': <class 'cfdm.data.gatheredarray.GatheredArray'>,
     'H5netcdfArray': <class 'cfdm.data.h5netcdfarray.H5netcdfArray'>,
     'NetCDF4Array': <class 'cfdm.data.netcdf4array.NetCDF4Array'>,
     'PointTopologyArray': <class 'cfdm.data.pointtopologyarray.PointTopologyArray'>,
     'RaggedContiguousArray': <class 'cfdm.data.raggedcontiguousarray.RaggedContiguousArray'>,
     'RaggedIndexedArray': <class 'cfdm.data.raggedindexedarray.RaggedIndexedArray'>,
     'RaggedIndexedContiguousArray': <class 'cfdm.data.raggedindexedcontiguousarray.RaggedIndexedContiguousArray'>,
     'SubsampledArray': <class 'cfdm.data.subsampledrray.SubsampledArray'>,
     'List': <class 'cfdm.list.List'>,
     'Count': <class 'cfdm.count.Count'>,
     'Index': <class 'cfdm.index.Index'>,
     'NodeCountProperties': <class 'cfdm.nodecountproperties.NodeCountProperties'>,
     'PartNodeCountProperties': <class 'cfdm.partnodecountproperties.PartNodeCountProperties'>}

    """
    return _implementation.copy()



from collections.abc import Iterable

from ...cfdmimplementation import implementation
from ...core import DocstringRewriteMeta
from ...docstring import _docstring_substitution_definitions


class ReadWrite(metaclass=DocstringRewriteMeta):
    """TODOCFA."""

    implementation = implementation()

    def __docstring_substitutions__(self):
        """Defines applicable docstring substitutions.

        Substitutons are considered applicable if they apply to this
        class as well as all of its subclasses.

        These are in addtion to, and take precendence over, docstring
        substitutions defined by the base classes of this class.

        See `_docstring_substitutions` for details.

        .. versionaddedd:: (cfdm) 1.12.0.0

        :Returns:

            `dict`
                The docstring substitutions that have been applied.

        """
        return _docstring_substitution_definitions

    def __docstring_package_depth__(self):
        """Returns the package depth for {{package}} substitutions.

        See `_docstring_package_depth` for details.

        .. versionaddedd:: (cfdm) 1.12.0.0

        """
        return 0

    @classmethod
    def _flat(cls, x):
        """Return an iterator over an arbitrarily nested sequence.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            x: scalar or arbitrarily nested sequence
                The arbitrarily nested sequence to be flattened. Note
                that if *x* is a string or a scalar then this is
                equivalent to passing a single element sequence
                containing *x*.

        :Returns:

            generator
                An iterator over the flattened sequence.

        **Examples**

        >>> list({{package}}.write._flat([1, (2, [3, 4])]))
        [1, 2, 3, 4]

        >>> list({{package}}.write._flat(['a', ['bc', ['def', 'ghij']]]))
        ['a', 'bc', 'def', 'ghij']

        >>> list({{package}}.write._flat(2004))
        [2004]

        """
        if not isinstance(x, Iterable) or isinstance(x, str):
            x = (x,)

        for a in x:
            if not isinstance(a, str) and isinstance(a, Iterable):
                for sub in cls._flat(a):
                    yield sub
            else:
                yield a



import atexit
import datetime
import faulthandler
import os
import tempfile
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

n_tmpfiles = 1
tmpfiles = [
    tempfile.mkstemp("_test_CoordinateReference.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(tempfile1,) = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class CoordinateReferenceTest(unittest.TestCase):
    """Unit test for the CoordinateReference class."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.LOG_LEVEL("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

        self.filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
        )
        f = cfdm.read(self.filename)
        self.assertEqual(len(f), 1, f"f={f!r}")
        self.f = f[0]

    def test_CoordinateReference__repr__str__dump_construct_type(self):
        """Test all means of CoordinateReference inspection."""
        f = self.f

        for cr in f.coordinate_references().values():
            _ = repr(cr)
            _ = str(cr)
            self.assertIsInstance(cr.dump(display=False), str)
            self.assertEqual(cr.construct_type, "coordinate_reference")

    def test_CoordinateReference_equals(self):
        """Test the equality-testing CoordinateReference method."""
        # Create a vertical grid mapping coordinate reference
        t = cfdm.CoordinateReference(
            coordinates=("coord1",),
            coordinate_conversion=cfdm.CoordinateConversion(
                parameters={
                    "standard_name": "atmosphere_hybrid_height_coordinate"
                },
                domain_ancillaries={"a": "aux0", "b": "aux1", "orog": "orog"},
            ),
        )
        self.assertTrue(t.equals(t.copy(), verbose=3))

        # Create a horizontal grid mapping coordinate reference
        t = cfdm.CoordinateReference(
            coordinates=["coord1", "fred", "coord3"],
            coordinate_conversion=cfdm.CoordinateConversion(
                parameters={
                    "grid_mapping_name": "rotated_latitude_longitude",
                    "grid_north_pole_latitude": 38.0,
                    "grid_north_pole_longitude": 190.0,
                }
            ),
        )
        self.assertTrue(t.equals(t.copy(), verbose=3))

        datum = cfdm.Datum(parameters={"earth_radius": 6371007})
        conversion = cfdm.CoordinateConversion(
            parameters={
                "grid_mapping_name": "rotated_latitude_longitude",
                "grid_north_pole_latitude": 38.0,
                "grid_north_pole_longitude": 190.0,
            }
        )

        t = cfdm.CoordinateReference(
            coordinate_conversion=conversion,
            datum=datum,
            coordinates=["x", "y", "lat", "lon"],
        )

        self.assertTrue(t.equals(t.copy(), verbose=3))

        # Create a horizontal grid mapping coordinate reference
        t = cfdm.CoordinateReference(
            coordinates=["coord1", "fred", "coord3"],
            coordinate_conversion=cfdm.CoordinateConversion(
                parameters={
                    "grid_mapping_name": "albers_conical_equal_area",
                    "standard_parallel": [-30, 10],
                    "longitude_of_projection_origin": 34.8,
                    "false_easting": -20000,
                    "false_northing": -30000,
                }
            ),
        )
        self.assertTrue(t.equals(t.copy(), verbose=3))

        # Create a horizontal grid mapping coordinate reference
        t = cfdm.CoordinateReference(
            coordinates=["coord1", "fred", "coord3"],
            coordinate_conversion=cfdm.CoordinateConversion(
                parameters={
                    "grid_mapping_name": "albers_conical_equal_area",
                    "standard_parallel": cfdm.Data([-30, 10]),
                    "longitude_of_projection_origin": 34.8,
                    "false_easting": -20000,
                    "false_northing": -30000,
                }
            ),
        )
        self.assertTrue(t.equals(t.copy(), verbose=3))

    def test_CoordinateConversion(self):
        """Test a CoordinateReference coordinate conversion element."""
        f = self.f.copy()

        cr = f.construct("standard_name:atmosphere_hybrid_height_coordinate")
        cc = cr.coordinate_conversion
        _ = repr(cc)
        _ = str(cc)

        domain_ancillaries = cc.domain_ancillaries()
        self.assertEqual(len(domain_ancillaries), 3)

        for key, value in domain_ancillaries.items():
            self.assertTrue(cc.has_domain_ancillary(key))
            self.assertEqual(cc.get_domain_ancillary(key), value)
            _ = cc.del_domain_ancillary(key)
            self.assertFalse(cc.has_domain_ancillary(key))
            self.assertIsNone(cc.get_domain_ancillary(key, None))
            self.assertIsNone(cc.del_domain_ancillary(key, None))
            cc.set_domain_ancillary(key, _)
            self.assertTrue(cc.has_domain_ancillary(key))
            self.assertEqual(cc.get_domain_ancillary(key), value)

        cr = f.construct("grid_mapping_name:rotated_latitude_longitude")
        cc = cr.coordinate_conversion
        _ = repr(cc)
        _ = str(cc)

        parameters = cc.parameters()
        self.assertEqual(len(parameters), 3, parameters)

        for key, value in parameters.items():
            self.assertTrue(cc.has_parameter(key))
            self.assertEqual(cc.get_parameter(key), value)
            _ = cc.del_parameter(key)
            self.assertFalse(cc.has_parameter(key))
            self.assertIsNone(cc.get_parameter(key, None))
            self.assertIsNone(cc.del_parameter(key, None))
            cc.set_parameter(key, _)
            self.assertTrue(cc.has_parameter(key))
            self.assertEqual(cc.get_parameter(key), value)

        _ = cr.del_coordinate_conversion()
        self.assertTrue(_.equals(cc))

    def test_Datum(self):
        """Test a CoordinateReference datum component."""
        f = self.f.copy()

        cr = f.construct("standard_name:atmosphere_hybrid_height_coordinate")
        d = cr.datum
        _ = repr(d)
        _ = str(d)

        parameters = d.parameters()
        self.assertEqual(len(parameters), 1, parameters)

        for key, value in parameters.items():
            self.assertTrue(d.has_parameter(key))
            self.assertEqual(d.get_parameter(key), value)
            _ = d.del_parameter(key)
            self.assertFalse(d.has_parameter(key))
            self.assertIsNone(d.get_parameter(key, None))
            self.assertIsNone(d.del_parameter(key, None))
            d.set_parameter(key, _)
            self.assertTrue(d.has_parameter(key))
            self.assertEqual(d.get_parameter(key), value)

        _ = cr.del_datum()
        self.assertTrue(_.equals(d))

        f = self.f.copy()
        key = f.construct_key("grid_mapping_name:rotated_latitude_longitude")
        f.del_construct(key)
        cr = f.construct("standard_name:atmosphere_hybrid_height_coordinate")
        cr.datum.nc_set_variable("my_name")
        cfdm.write(f, tempfile1)

    def test_CoordinateReference_parameters(self):
        """Test the parameter access methods of CoordinateReference."""
        f = self.f.copy()

        cr = f.construct("grid_mapping_name:rotated_latitude_longitude")

        cr.datum.set_parameter("test", 999)
        cr.coordinate_conversion.set_parameter("test", 111)

        with self.assertRaises(ValueError):
            cfdm.write(f, tempfile1)

    def test_CoordinateReference_write(self):
        """Test write when vertical CRS has no coordinates."""
        f = cfdm.example_field(1)
        c = f.construct("standard_name:atmosphere_hybrid_height_coordinate")
        c.clear_coordinates()
        # This write should not fail ...
        cfdm.write(f, tempfile1)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



import atexit
import datetime
import faulthandler
import os
import tempfile
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

n_tmpfiles = 3
tmpfiles = [
    tempfile.mkstemp("_test_external.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(tempfile, tempfile_parent, tempfile_external) = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class ExternalVariableTest(unittest.TestCase):
    """Test handling of variables stored in external netCDF files."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.LOG_LEVEL("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

        dataset_dir = os.path.dirname(os.path.abspath(__file__))
        self.parent_file = os.path.join(dataset_dir, "parent.nc")
        self.external_file = os.path.join(dataset_dir, "external.nc")
        self.combined_file = os.path.join(dataset_dir, "combined.nc")
        self.external_missing_file = os.path.join(
            dataset_dir, "external_missing.nc"
        )

    def test_EXTERNAL_READ(self):
        """Test read function reading of external netCDF variables."""
        # Read the parent file on its own, without the external file
        f = cfdm.read(self.parent_file)

        for i in f:
            _ = repr(i)
            _ = str(i)
            self.assertIsInstance(i.dump(display=False), str)

        self.assertEqual(len(f), 1)
        f = f[0]

        cell_measure = f.constructs.filter_by_identity("measure:area").value()

        self.assertTrue(cell_measure.nc_get_external())
        self.assertEqual(cell_measure.nc_get_variable(), "areacella")
        self.assertEqual(cell_measure.properties(), {})
        self.assertFalse(cell_measure.has_data())

        # External file contains only the cell measure variable
        f = cfdm.read(
            self.parent_file, external=[self.external_file], verbose=False
        )

        c = cfdm.read(self.combined_file, verbose=False)

        for i in c + f:
            _ = repr(i)
            _ = str(i)
            self.assertIsInstance(i.dump(display=False), str)

        cell_measure = (
            f[0].constructs.filter_by_identity("measure:area").value()
        )

        self.assertEqual(len(f), 1)
        self.assertEqual(len(c), 1)

        for i in range(len(f)):
            self.assertTrue(c[i].equals(f[i], verbose=3))

        # External file contains other variables
        f = cfdm.read(
            self.parent_file, external=self.combined_file, verbose=False
        )

        for i in f:
            _ = repr(i)
            _ = str(i)
            self.assertIsInstance(i.dump(display=False), str)

        self.assertEqual(len(f), 1)
        self.assertEqual(len(c), 1)

        for i in range(len(f)):
            self.assertTrue(c[i].equals(f[i], verbose=3))

        # Two external files
        f = cfdm.read(
            self.parent_file,
            external=[self.external_file, self.external_missing_file],
            verbose=False,
        )

        for i in f:
            _ = repr(i)
            _ = str(i)
            self.assertIsInstance(i.dump(display=False), str)

        self.assertEqual(len(f), 1)
        self.assertEqual(len(c), 1)

        for i in range(len(f)):
            self.assertTrue(c[i].equals(f[i], verbose=3))

    def test_EXTERNAL_WRITE(self):
        """Test write function writing of external netCDF variables."""
        parent = cfdm.read(self.parent_file)
        combined = cfdm.read(self.combined_file)

        # External file contains only the cell measure variable
        f = cfdm.read(self.parent_file, external=self.external_file)

        cfdm.write(f, tempfile)
        g = cfdm.read(tempfile)

        self.assertEqual(len(g), len(combined))

        for i in range(len(g)):
            self.assertTrue(combined[i].equals(g[i], verbose=3))

        cell_measure = g[0].constructs("measure:area").value()

        self.assertFalse(cell_measure.nc_get_external())
        cell_measure.nc_set_external(True)
        self.assertTrue(cell_measure.nc_get_external())
        self.assertTrue(cell_measure.properties())
        self.assertTrue(cell_measure.has_data())

        self.assertTrue(
            g[0]
            .constructs.filter_by_identity("measure:area")
            .value()
            .nc_get_external()
        )

        cfdm.write(
            g, tempfile_parent, external=tempfile_external, verbose=False
        )

        h = cfdm.read(tempfile_parent, verbose=False)

        self.assertEqual(len(h), len(parent))

        for i in range(len(h)):
            self.assertTrue(parent[i].equals(h[i], verbose=3))

        h = cfdm.read(tempfile_external)
        external = cfdm.read(self.external_file)

        self.assertEqual(len(h), len(external))

        for i in range(len(h)):
            self.assertTrue(external[i].equals(h[i], verbose=3))


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import faulthandler
import os
import unittest

import cftime
import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class DimensionCoordinateTest(unittest.TestCase):
    """Unit test for the DimensionCoordinate class."""

    filename = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
    )

    dim = cfdm.DimensionCoordinate()
    dim.set_property("standard_name", "latitude")
    a = np.array(
        [
            -30,
            -23.5,
            -17.8123,
            -11.3345,
            -0.7,
            -0.2,
            0,
            0.2,
            0.7,
            11.30003,
            17.8678678,
            23.5,
            30,
        ]
    )
    dim.set_data(cfdm.Data(a, "degrees_north"))
    bounds = cfdm.Bounds()
    b = np.empty(a.shape + (2,))
    b[:, 0] = a - 0.1
    b[:, 1] = a + 0.1
    bounds.set_data(cfdm.Data(b))
    dim.set_bounds(bounds)

    f0 = cfdm.example_field(0)

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_DimensionCoordinate__repr__str__dump(self):
        """Test all means of DimensionCoordinate inspection."""
        f = cfdm.read(self.filename)[0]
        x = f.dimension_coordinates("grid_latitude").value()

        _ = repr(x)
        _ = str(x)
        self.assertIsInstance(x.dump(display=False), str)
        self.assertIsInstance(x.dump(display=False, _key="qwerty"), str)

    def test_DimensionCoordinate__init__(self):
        """Test the constructor of DimensionCoordinate."""
        cfdm.DimensionCoordinate(source="qwerty")

    def test_DimensionCoordinate_set_data(self):
        """Test the `set_data` DimensionCoordinate method."""
        x = cfdm.DimensionCoordinate()

        y = x.set_data(cfdm.Data([1, 2, 3]))
        self.assertIsNone(y)
        self.assertTrue(x.has_data())

        # Test inplace
        x.del_data()
        y = x.set_data(cfdm.Data([1, 2, 3]), inplace=False)
        self.assertIsInstance(y, cfdm.DimensionCoordinate)
        self.assertFalse(x.has_data())
        self.assertTrue(y.has_data())

        # Exceptions should be raised for 0-d and N-d (N>=2) data
        with self.assertRaises(Exception):
            y = x.set_data(cfdm.Data([[1, 2, 3]]))

        with self.assertRaises(Exception):
            y = x.set_data(cfdm.Data(1))

    def test_DimensionCoordinate_climatology(self):
        """Test the climatology DimensionCoordinate methods."""
        x = cfdm.DimensionCoordinate()

        self.assertFalse(x.is_climatology())
        self.assertIsNone(x.get_climatology(None))
        x.set_climatology(False)
        self.assertFalse(x.is_climatology())
        self.assertFalse(x.get_climatology())
        x.set_climatology(True)
        self.assertTrue(x.is_climatology())
        self.assertTrue(x.del_climatology())
        self.assertIsNone(x.del_climatology(None))

    def test_DimensiconCoordinate_array(self):
        """Test the `DimensionCoordinate.array` method."""
        f = self.f0
        t = f.construct("time")
        self.assertEqual(t.array, 31)
        self.assertEqual(t.array, t.data.array)

    def test_DimensiconCoordinate_datetime_array(self):
        """Test the `DimensionCoordinate.datetime_array` method."""
        f = self.f0
        t = f.construct("time")
        self.assertEqual(
            t.datetime_array,
            np.array([cftime.DatetimeGregorian(2019, 1, 1, 0, 0, 0, 0)]),
        )
        self.assertEqual(t.datetime_array, t.data.datetime_array)

    def test_DimensiconCoordinate__getitem__(self):
        """Test the `DimensionCoordinate.__getitem__` method."""
        dim = self.dim
        self.assertTrue((dim[1:3].array == dim.array[1:3]).all())

        # Indices result in a subspaced shape that has a size 0 axis
        with self.assertRaises(IndexError):
            dim[[False] * dim.size]

    def test_DimensionCoordinate_hdf5_chunksizes(self):
        """Test the DimensionCoordinate HDF5 chunksizes methods."""
        c = self.dim.copy()

        self.assertIsNone(c.nc_set_hdf5_chunksizes([4]))
        self.assertEqual(c.nc_hdf5_chunksizes(), (4,))
        self.assertEqual(c.bounds.nc_hdf5_chunksizes(), (4, 2))

        c.nc_set_hdf5_chunksizes(1024, bounds=False, interior_ring=False)
        self.assertEqual(c.nc_hdf5_chunksizes(), 1024)
        self.assertEqual(c.bounds.nc_hdf5_chunksizes(), (4, 2))

        c.nc_clear_hdf5_chunksizes(bounds=False, interior_ring=False)
        self.assertIsNone(c.nc_hdf5_chunksizes())
        self.assertEqual(c.bounds.nc_hdf5_chunksizes(), (4, 2))

        c.nc_clear_hdf5_chunksizes()
        self.assertIsNone(c.nc_hdf5_chunksizes())
        self.assertIsNone(c.bounds.nc_hdf5_chunksizes())


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



from . import core, mixin


class DimensionCoordinate(
    mixin.NetCDFVariable,
    mixin.Coordinate,
    mixin.Files,
    core.DimensionCoordinate,
):
    """A dimension coordinate construct of the CF data model.

    A dimension coordinate construct provides information which locate
    the cells of the domain and which depend on a subset of the domain
    axis constructs. The dimension coordinate construct is able to
    unambiguously describe cell locations because a domain axis can be
    associated with at most one dimension coordinate construct, whose
    data array values must all be non-missing and strictly
    monotonically increasing or decreasing. They must also all be of
    the same numeric data type. If cell bounds are provided, then each
    cell must have exactly two vertices. CF-netCDF coordinate
    variables and numeric scalar coordinate variables correspond to
    dimension coordinate constructs.

    The dimension coordinate construct consists of a data array of the
    coordinate values which spans a subset of the domain axis
    constructs, an optional array of cell bounds recording the extents
    of each cell (stored in a `Bounds` object), and properties to
    describe the coordinates. An array of cell bounds spans the same
    domain axes as its coordinate array, with the addition of an extra
    dimension whose size is that of the number of vertices of each
    cell. This extra dimension does not correspond to a domain axis
    construct since it does not relate to an independent axis of the
    domain. Note that, for climatological time axes, the bounds are
    interpreted in a special way indicated by the cell method
    constructs.

    **NetCDF interface**

    {{netCDF variable}}

    {{netCDF HDF5 chunks}}

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        properties=None,
        data=None,
        bounds=None,
        geometry=None,
        interior_ring=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

               *Parameter example:*
                  ``properties={'standard_name': 'time'}``

            {{init data: data_like, optional}}

            {{init bounds: `Bounds`, optional}}

            {{init geometry: `str`, optional}}

            {{init interior_ring: `InteriorRing`, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            properties=properties,
            data=data,
            bounds=bounds,
            geometry=geometry,
            interior_ring=interior_ring,
            source=source,
            copy=copy,
            _use_data=_use_data,
        )

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

    def dump(
        self,
        display=True,
        _omit_properties=None,
        _key=None,
        _level=0,
        _title=None,
        _axes=None,
        _axis_names=None,
    ):
        """A full description of the dimension coordinate construct.

        Returns a description of all properties, including those of
        components, and provides selected values of all data arrays.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        if _title is None:
            if _key is None:
                default = ""
            else:
                default = f"key%{_key}"

            _title = "Dimension coordinate: " + self.identity(default=default)

        return super().dump(
            display=display,
            _omit_properties=_omit_properties,
            _key=_key,
            _level=_level,
            _title=_title,
            _axes=_axes,
            _axis_names=_axis_names,
        )



"""Create a flattened version of a grouped netCDF dataset.

Portions of this code were adapted from the `netcdf_flattener`
library, which carries the following Apache 2.0 Licence:

Copyright (c) 2020 EUMETSAT

Licensed to the Apache Software Foundation (ASF) under one or more
contributor license agreements. The ASF licenses this file to you
under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy
of the License at http://www.apache.org/licenses/LICENSE-2.0.

"""

import hashlib
import logging
import re
import warnings

from .config import (
    flattener_attribute_map,
    flattener_dimension_map,
    flattener_separator,
    flattener_variable_map,
    flattening_rules,
    group_separator,
    max_name_len,
    ref_not_found_error,
)

# Mapping from numpy dtype endian format to that expected by netCDF4
_dtype_endian_lookup = {
    "=": "native",
    ">": "big",
    "<": "little",
    "|": "native",
    None: "native",
}

# Set of netCDF attributes that contain references to dimensions or
# variables
referencing_attributes = set(flattening_rules)


def netcdf_flatten(
    input_ds,
    output_ds,
    strict=True,
    omit_data=False,
    write_chunksize=134217728,
):
    """Create a flattened version of a grouped netCDF dataset.

    **CF-netCDF coordinate variables**

    When a CF-netCDF coordinate variable in the input dataset is in a
    different group to its corresponding dimension, the same variable
    in the output flattened dataset will no longer be a CF-netCDF
    coordinate variable, as its name will be prefixed with a different
    group identifier than its dimension.

    In such cases it is up to the user to apply the proximal and
    lateral search algorithms to the flattened dataset returned by
    `netcdf_flatten`, in conjunction with the mappings defined in the
    newly created global attributes ``_flattener_variable_map`` and
    ``_flattener_dimension_map``, to find which netCDF variables are
    acting as CF coordinate variables in the flattened dataset. See
    https://cfconventions.org/cf-conventions/cf-conventions.html#groups
    for details.

    For example, if an input dataset has dimension ``lat`` in the root
    group and coordinate variable ``lat(lat)`` in group ``/group1``,
    then the flattened dataset will contain dimension ``lat`` and
    variable ``group1__lat(lat)``, both in its root group. In this
    case, the ``_flattener_variable_map`` global attribute of the
    flattened dataset will contain the mapping ``'group1__lat:
    /group1/lat'``, and the ``_flattener_dimension_map`` global
    attribute will contain the mapping ``'lat: /lat'``.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        input_ds:
            The dataset to be flattened, that has the same API as
            `netCDF4.Dataset` or `h5netcdf.File`.

        output_ds: `netCDF4.Dataset`
            A container for the flattened dataset.

        strict: `bool`, optional
            If True, the default, then failing to resolve a reference
            raises an exception. If False, a warning is issued and
            flattening is continued.

        omit_data: `bool`, optional
            If True then do not copy the data of any variables from
            *input_ds* to *output_ds*. This does not affect the amount
            of netCDF variables and dimensions that are written to the
            file, nor the netCDF variables' attributes, but for all
            variables it does not create data on disk or in
            memory. The resulting dataset will be smaller than it
            otherwise would have been, and when the new dataset is
            accessed the data of these variables will be represented
            by an array of all missing data. If False, the default,
            then all data arrays are copied.

        write_chunksize: `int`, optional
            When *omit_data* is False, the copying of data is done
            piecewise to keep memory usage down. *write_chunksize* is
            the size in bytes of how much data is copied from
            *input_ds* to *output_ds* for each piece. Ignored if
            *omit_data* is True.

    """
    _Flattener(
        input_ds,
        output_ds,
        strict,
        omit_data=omit_data,
        write_chunksize=write_chunksize,
    ).flatten()


def parse_attribute(name, attribute):
    """Parse variable attribute of any form into a dict:

     * 'time' -> {'time': []}
     * 'lat lon' -> {'lat': [], 'lon': []}
     * 'area: time volume: lat lon' -> {'area': ['time'], 'volume':
       ['lat', 'lon']}

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        name: `str`
            The attribute name (e.g. ``'cell_methods'``).

        attribute: `str`
            The attribute value to parse.

    :Returns:

        `dict`
            The parsed string.

    """

    def subst(s):
        """Substitute tokens for WORD and SEP."""
        return s.replace("WORD", r"[A-Za-z0-9_#/.\(\)]+").replace(
            "SEP", r"(\s+|$)"
        )

    # Regex for 'dict form': "k1: v1 v2 k2: v3"
    pat_value = subst(r"(?P<value>WORD)SEP")
    pat_values = f"({pat_value})*"
    pat_mapping = subst(rf"(?P<mapping_name>WORD):SEP(?P<values>{pat_values})")
    pat_mapping_list = f"({pat_mapping})+"

    # Regex for 'list form': "v1 v2 v3" (including single-item form)
    pat_list_item = subst(r"(?P<list_item>WORD)SEP")
    pat_list = f"({pat_list_item})+"

    # Regex for any form:
    pat_all = subst(
        rf"((?P<list>{pat_list})|(?P<mapping_list>{pat_mapping_list}))$"
    )

    m = re.match(pat_all, attribute)

    # Output is always a dict. If input form is a list, dict values
    # are set as empty lists
    out = {}

    if m is not None:
        list_match = m.group("list")
        # Parse as a list
        if list_match:
            for mapping in re.finditer(pat_list_item, list_match):
                item = mapping.group("list_item")
                out[item] = None

        # Parse as a dict:
        else:
            mapping_list = m.group("mapping_list")
            for mapping in re.finditer(pat_mapping, mapping_list):
                term = mapping.group("mapping_name")
                values = [
                    value.group("value")
                    for value in re.finditer(
                        pat_value, mapping.group("values")
                    )
                ]
                out[term] = values
    else:
        raise AttributeParsingException(
            f"Error parsing {name!r} attribute with value {attribute!r}"
        )

    return out


def generate_var_attr_str(d):
    """Re-generate the attribute string from a dictionary.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        d: `dict`
            A resolved and parsed attribute.

    :Returns:

        `str`
            The flattened attribute value.

    """
    parsed_list = []
    for k, v in d.items():
        if v is None:
            parsed_list.append(k)
        elif not v:
            parsed_list.append(f"{k}:")
        else:
            parsed_list.append(f"{k}: {' '.join(v)}")

    return " ".join(parsed_list)


class _Flattener:
    """Information and methods needed to flatten a netCDF dataset.

    Contains the input file, the output file being flattened, and all
    the logic of the flattening process.

    .. versionadded:: (cfdm) 1.11.2.0

    """

    def __init__(
        self,
        input_ds,
        output_ds,
        strict=True,
        omit_data=False,
        write_chunksize=134217728,
    ):
        """**Initialisation**

        :Parameters:

            input_ds:
                The dataset to be flattened, that has the same API as
                `netCDF4.Dataset` or `h5netcdf.File`.

            output_ds: `netCDF4.Dataset`
                A container for the flattened dataset.

            strict: `bool`, optional
                See `netcdf_flatten`.

            omit_data: `bool`, optional
                See `netcdf_flatten`.

            write_chunksize: `int`, optional
                See `netcdf_flatten`.

        """
        self._attr_map_value = []
        self._dim_map_value = []
        self._var_map_value = []

        self._dim_map = {}
        self._var_map = {}

        self._input_ds = input_ds
        self._output_ds = output_ds
        self._strict = bool(strict)
        self._omit_data = bool(omit_data)
        self._write_chunksize = write_chunksize

        if (
            output_ds == input_ds
            or output_ds.filepath() == self.filepath(input_ds)
            or output_ds.data_model != "NETCDF4"
        ):
            raise ValueError(
                "Invalid inputs. Input and output datasets should "
                "be different, and output should be of the 'NETCDF4' format."
            )

    def attrs(self, variable):
        """Return the variable attributes.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            variable:
                The variable, that has the same API as
                `netCDF4.Variable` or `h5netcdf.Variable`.

        :Returns:

            `dict`
                A dictionary of the attribute values keyed by their
                names.

        """
        try:
            # h5netcdf
            return dict(variable.attrs)
        except AttributeError:
            # netCDF4
            return {
                attr: variable.getncattr(attr) for attr in variable.ncattrs()
            }

    def chunksizes(self, variable):
        """Return the variable chunk sizes.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            variable:
                The variable, that has the same API as
                `netCDF4.Variable` or `h5netcdf.Variable`.

        :Returns:

            `None` or sequence of `int`
                The chunksizes, or `None` if the variable is not
                chunked.

        **Examples**

        >>> f.chunksizes(variable)
        [1, 324, 432]

        >>> f.chunksizes(variable)
        None

        """
        try:
            # netCDF4
            chunking = variable.chunking()
            if chunking == "contiguous":
                return None

            return chunking
        except AttributeError:
            # h5netcdf
            return variable.chunks

    def contiguous(self, variable):
        """Whether or not the variable data is contiguous on disk.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            variable:
                The variable, that has the same API as
                `netCDF4.Variable` or `h5netcdf.Variable`.

        :Returns:

            `bool`
                 `True` if the variable data is contiguous on disk,
                 otherwise `False`.

        **Examples**

        >>> f.contiguous(variable)
        False

        """
        try:
            # netCDF4
            return variable.chunking() == "contiguous"
        except AttributeError:
            # h5netcdf
            return variable.chunks is None

    def dtype(self, variable):
        """Return the data type of a variable.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            variable:
                The variable, that has the same API as
                `netCDF4.Variable` or `h5netcdf.Variable`.

        :Returns:

            `numpy.dtype` or `str`
                 The data type.

        **Examples**

        >>> f.dtype(variable)
        dtype('<f8')

        >>> f.dtype(variable)
        str

        """
        out = variable.dtype
        if out == "O":
            out = str

        return out

    def endian(self, variable):
        """Return the endian-ness of a variable.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            variable:
                The variable, that has the same API as
                `netCDF4.Variable` or `h5netcdf.Variable`.

        :Returns:

            `str`
                 The endian-ness (``'little'``, ``'big'``, or
                 ``'native'``) of the variable.

        **Examples**

        >>> f.endian(variable)
        'native'

        """
        try:
            # netCDF4
            return variable.endian()
        except AttributeError:
            # h5netcdf
            dtype = variable.dtype
            return _dtype_endian_lookup[getattr(dtype, "byteorder", None)]

    def filepath(self, dataset):
        """Return the file path for the dataset.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            dataset:
                The dataset, that has the same API as
                `netCDF4.Dataset` or `h5netcdf.File`.

        :Returns:

            `str`
                The file system path, or the OPeNDAP URL, for the
                dataset.

        **Examples**

        >>> f.filepath(dataset)
        '/home/data/file.nc'

        """
        try:
            # netCDF4
            return dataset.filepath()
        except AttributeError:
            # h5netcdf
            return dataset.filename

    def get_dims(self, variable):
        """Return the dimensions associated with a variable.

        .. versionadded:: (cfdm) 1.11.2.0

        :Returns:

            `list`

        """
        try:
            # netCDF4
            return variable.get_dims()
        except AttributeError:
            # h5netcdf
            dims = {}
            dimension_names = list(variable.dimensions)
            group = variable._parent
            for name, dim in group.dims.items():
                if name in dimension_names:
                    dims[name] = dim
                    dimension_names.remove(name)

            group = group.parent
            while group is not None and dimension_names:
                for name, dim in group.dims.items():
                    if name in dimension_names:
                        dims[name] = dim
                        dimension_names.remove(name)

                group = group.parent

            return [dims[name] for name in variable.dimensions]

    def getncattr(self, x, attr):
        """Retrieve a netCDF attribute.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            x: variable, group, or dataset

            attr: `str`

        :Returns:

        """
        try:
            # netCDF4
            return getattr(x, attr)
        except AttributeError:
            # h5netcdf
            return x.attrs[attr]

    def group(self, x):
        """Return the group that a variable belongs to.

        .. versionadded:: (cfdm) 1.11.2.0

        :Returns:

            `Group`

        """
        try:
            # netCDF4
            return x.group()
        except AttributeError:
            # h5netcdf
            return x._parent

    def name(self, x):
        """Return the netCDF name, without its groups.

        .. versionadded:: (cfdm) 1.11.2.0

        :Returns:

            `str`

        """
        out = x.name
        if group_separator in out:
            # h5netcdf
            out = x.name.split(group_separator)[-1]

        return out

    def ncattrs(self, x):
        """Return netCDF attribute names.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            x: variable, group, or dataset

        :Returns:

            `list`

        """
        try:
            # netCDF4
            return x.ncattrs()
        except AttributeError:
            # h5netcdf
            return list(x.attrs)

    def parent(self, group):
        """Return a simulated unix parent group.

        .. versionadded:: (cfdm) 1.11.2.0

        :Returns:

            `str`

        """
        try:
            return group.parent
        except AttributeError:
            return

    def path(self, group):
        """Return a simulated unix directory path to a group.

        .. versionadded:: (cfdm) 1.11.2.0

        :Returns:

            `str`

        """
        try:
            # netCDF4
            return group.path
        except AttributeError:
            # h5netcdf
            try:
                return group.name
            except AttributeError:
                return group_separator

    def flatten(self):
        """Flattens and writes to output file.

        .. versionadded:: (cfdm) 1.11.2.0

        :Return:

            `None`

        """
        input_ds = self._input_ds
        output_ds = self._output_ds

        logging.info(f"Flattening the groups of {self.filepath(input_ds)}")

        # Flatten product
        self.process_group(input_ds)

        # Add name mapping attributes
        output_ds.setncattr(flattener_attribute_map, self._attr_map_value)
        output_ds.setncattr(flattener_dimension_map, self._dim_map_value)
        output_ds.setncattr(flattener_variable_map, self._var_map_value)

        # Browse flattened variables to rename references:
        logging.info(
            "    Browsing flattened variables to rename references "
            "in attributes"
        )
        for var in output_ds.variables.values():
            self.adapt_references(var)

    def process_group(self, input_group):
        """Flattens a given group to the output file.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            input_group: `str`
                The group to flatten.

        :Returns:

            `None`

        """
        logging.info(f"    Browsing group {self.path(input_group)}")

        for attr_name in self.ncattrs(input_group):
            self.flatten_attribute(input_group, attr_name)

        for dim in input_group.dimensions.values():
            self.flatten_dimension(dim)

        for var in input_group.variables.values():
            self.flatten_variable(var)

        for child_group in input_group.groups.values():
            self.process_group(child_group)

    def flatten_attribute(self, input_group, attr_name):
        """Flattens a given attribute from a group to the output file.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            input_group: `str`
                The group containing the attribute to flatten.

            attr_name: `str`
                The name of the attribute.

        :Returns:

            `None`

        """
        logging.info(
            f"        Copying attribute {attr_name} from "
            f"group {self.path(input_group)} to root"
        )

        # Create new name
        new_attr_name = self.generate_flattened_name(input_group, attr_name)

        # Write attribute
        self._output_ds.setncattr(
            new_attr_name, self.getncattr(input_group, attr_name)
        )

        # Store new naming for later and in mapping attribute
        self._attr_map_value.append(
            self.generate_mapping_str(input_group, attr_name, new_attr_name)
        )

    def flatten_dimension(self, dim):
        """Flattens a given dimension to the output file.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            dim:
                The dimension to flatten, that has the same API as
                `netCDF4.Dimension` or `h5netcdf.Dimension`.

        :Returns:

            `None`

        """
        logging.info(
            f"    Copying dimension {self.name(dim)} from "
            f"group {self.path(self.group(dim))} to root"
        )

        # Create new name
        new_name = self.generate_flattened_name(
            self.group(dim), self.name(dim)
        )

        # Write dimension
        self._output_ds.createDimension(
            new_name, (len(dim), None)[dim.isunlimited()]
        )

        # Store new name in dict for resolving references later
        self._dim_map[self.pathname(self.group(dim), self.name(dim))] = (
            new_name
        )

        # Add to name mapping attribute
        self._dim_map_value.append(
            self.generate_mapping_str(
                self.group(dim), self.name(dim), new_name
            )
        )

    def flatten_variable(self, var):
        """Flattens a given variable to the output file.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            var:
                The variable, that has the same API as
                `netCDF4.Variable` or `h5netcdf.Variable`.

        :Returns:

            `None`

        """
        logging.info(
            f"        Copying variable {self.name(var)} from "
            f"group {self.path(self.group(var))} to root"
        )

        # Create new name
        new_name = self.generate_flattened_name(
            self.group(var), self.name(var)
        )

        # Replace old by new dimension names
        new_dims = list(
            map(
                lambda x: self._dim_map[
                    self.pathname(self.group(x), self.name(x))
                ],
                self.get_dims(var),
            )
        )

        # Write variable
        fullname = self.pathname(self.group(var), self.name(var))
        logging.info(f"        Creating variable {new_name} from {fullname}")

        attributes = self.attrs(var)

        omit_data = self._omit_data
        if omit_data:
            fill_value = False
        else:
            fill_value = attributes.pop("_FillValue", None)

        new_var = self._output_ds.createVariable(
            new_name,
            self.dtype(var),
            new_dims,
            zlib=False,
            complevel=4,
            shuffle=True,
            fletcher32=False,
            contiguous=self.contiguous(var),
            chunksizes=self.chunksizes(var),
            endian=self.endian(var),
            least_significant_digit=None,
            fill_value=fill_value,
        )

        if not omit_data:
            self.write_data_in_chunks(var, new_var)

        # Copy attributes
        new_var.setncatts(attributes)

        # Store new name in dict for resolving references later
        self._var_map[self.pathname(self.group(var), self.name(var))] = (
            new_name
        )

        # Add to name mapping attribute
        self._var_map_value.append(
            self.generate_mapping_str(
                self.group(var), self.name(var), new_name
            )
        )

        # Resolve references in variable attributes and replace by
        # absolute path
        self.resolve_references(new_var, var)

    def increment_pos(self, pos, dim, copy_slice_shape, var_shape):
        """Increment position.

        Increment position vector in a variable along a dimension by
        the matching slice length along that dimension. If end of the
        dimension is reached, recursively increment the next
        dimensions until a valid position is found.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            pos: `list`
                The current slice position along each dimension of the
                array.

            dim: `int`
                The position of the array dimension to be incremented.

            copy_slice_shape: `list`
                The shape of the copy slice.

            var_shape: `tuple`
                The shape of the whole variable.

        :Returns:

            `bool`
                `True` if a valid position is found within the
                variable, `False` otherwise.

        """
        # Try to increment dimension
        pos[dim] += copy_slice_shape[dim]

        # Test new position
        dim_end_reached = pos[dim] > var_shape[dim]
        var_end_reached = (dim + 1) >= len(copy_slice_shape)

        # End of this dimension not reached yet
        if not dim_end_reached:
            return True

        # End of this dimension reached. Reset to 0 and try increment
        # next one recursively
        elif dim_end_reached and not var_end_reached:
            pos[: dim + 1] = [0 for j in range(dim + 1)]
            return self.increment_pos(
                pos, dim + 1, copy_slice_shape, var_shape
            )

        else:
            # End of this dimension reached, and no dimension to
            # increment. Finish.
            return False

    def write_data_in_chunks(self, old_var, new_var):
        """Copy the data of a variable to a new one by slice.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            old_var:
                The variable where the data should be copied from,
                that has the same API as `netCDF4.Variable` or
                `h5netcdf.Variable`.

            new_var:
                The new variable in which to copy the data, that has the
                same API as `netCDF4.Variable` or `h5netcdf.Variable`.

        :Returns:

            `None`

        """
        ndim = old_var.ndim
        shape = old_var.shape
        chunk_shape = (
            (self.write_chunksize // (old_var.dtype.itemsize * ndim)),
        ) * ndim

        logging.info(
            f"        Copying {self.name(old_var)!r} data in chunks of "
            f"{chunk_shape}"
        )
        # Initial position vector
        pos = [0] * ndim

        # Copy in slices until end reached
        var_end_reached = False
        while not var_end_reached:
            # Create current slice
            current_slice = tuple(
                slice(pos[dim_i], min(shape[dim_i], pos[dim_i] + dim_l))
                for dim_i, dim_l in enumerate(chunk_shape)
            )

            # Copy data in slice
            new_var[current_slice] = old_var[current_slice]

            # Get next position
            var_end_reached = not self.increment_pos(
                pos, 0, chunk_shape, shape
            )

    def resolve_reference(self, orig_ref, orig_var, rules):
        """Resolve a reference.

        Resolves the absolute path to a coordinate variable within the
        group structure.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            orig_ref: `str`
                The reference to resolve.

            orig_var:
                The original variable containing the reference, that
                has the same API as `netCDF4.Variable` or
                `h5netcdf.Variable`.

            rules: `FlatteningRules`
                The flattening rules that apply to the reference.

        :Returns:

            `str`
                The absolute path to the reference.

        """
        ref = orig_ref
        absolute_ref = None
        ref_type = ""

        ref_to_dim = rules.ref_to_dim
        ref_to_var = rules.ref_to_var

        # Resolve first as dim (True), or var (False)
        resolve_dim_or_var = ref_to_dim > ref_to_var

        # Resolve var (resp. dim) if resolving as dim (resp. var) failed
        resolve_alt = ref_to_dim and ref_to_var

        # Reference is already given by absolute path
        if ref.startswith(group_separator):
            method = "Absolute"
            absolute_ref = ref

        # Reference is given by relative path
        elif group_separator in ref:
            method = "Relative"

            # First tentative as dim OR var
            if resolve_dim_or_var:
                ref_type = "dimension"
            else:
                ref_type = "variable"

            absolute_ref = self.search_by_relative_path(
                orig_ref, self.group(orig_var), resolve_dim_or_var
            )

            # If failed and alternative possible, second tentative
            if absolute_ref is None and resolve_alt:
                if resolve_dim_or_var:
                    ref_type = "variable"
                else:
                    ref_type = "dimension"

                absolute_ref = self.search_by_relative_path(
                    orig_ref, self.groupp(orig_var), not resolve_dim_or_var
                )

        # Reference is to be searched by proximity
        else:
            method = "Proximity"
            absolute_ref, ref_type = self.resolve_reference_proximity(
                ref,
                resolve_dim_or_var,
                resolve_alt,
                orig_var,
                rules,
            )

        # Post-search checks and return result
        return self.resolve_reference_post_processing(
            absolute_ref,
            orig_ref,
            orig_var,
            rules,
            ref_type,
            method,
        )

    def resolve_reference_proximity(
        self, ref, resolve_dim_or_var, resolve_alt, orig_var, rules
    ):
        """Resolve reference: search by proximity.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            ref: `str`
                The reference to resolve.

            resolve_dim_or_var: `bool`
                Try to resolve first as dimension (True), or else as
                variable (False).

            resolve_alt: `bool`
                Resolve as variable if resolving as dimension failed,
                and vice versa.

            orig_var:
                The original variable containing the reference, that
                has the same API as `netCDF4.Variable` or
                `h5netcdf.Variable`.

            rules: `FlatteningRules`
                The flattening rules that apply to the reference.

        :Returns:

            (`str` or `None, str)
                The resolved reference (or `None` if unresolved), and
                the type of reference (either ``'dimension'`` or
                ``'variable'``).

        """
        # First tentative as dim OR var
        if resolve_dim_or_var:
            ref_type = "dimension"
        else:
            ref_type = "variable"

        stop_at_local_apex = rules.stop_at_local_apex

        resolved_var = self.search_by_proximity(
            ref,
            self.group(orig_var),
            resolve_dim_or_var,
            False,
            stop_at_local_apex,
        )

        # If failed and alternative possible, second tentative
        if resolved_var is None and resolve_alt:
            if resolve_dim_or_var:
                ref_type = "variable"
            else:
                ref_type = "dimension"

            resolved_var = self.search_by_proximity(
                ref,
                self.group(orig_var),
                not resolve_dim_or_var,
                False,
                stop_at_local_apex,
            )

        # If found, create ref string
        if resolved_var is not None:
            return (
                self.pathname(
                    self.group(resolved_var), self.name(resolved_var)
                ),
                ref_type,
            )
        else:
            return None, ""

    def resolve_reference_post_processing(
        self, absolute_ref, orig_ref, orig_var, rules, ref_type, method
    ):
        """Post-processing operations after resolving reference.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            absolute_ref: `str`
                The absolute path of the reference.

            orig_ref: `str`
                The original reference.

            orig_var:
                The original variable containing the reference, that
                has the same API as `netCDF4.Variable` or
                `h5netcdf.Variable`.

            rules: `FlatteningRules`
                The flattening rules that apply to the reference.

            ref_type: `str`
                the type of reference (either ``'dimension'`` or
                ``'variable'``).

            method: `str`
                The method of reference resolution (either
                ``'proximity'`` or ``'absolute'``).

        :Returns:

            `str`
                The absolute reference.

        """
        # If not found and accept standard name, assume standard name
        if absolute_ref is None and rules.accept_standard_names:
            logging.info(
                f"            Reference to {orig_ref!r} not "
                "resolved. Assumed to be a standard name."
            )
            ref_type = "standard_name"
            absolute_ref = orig_ref
        elif absolute_ref is None:
            # Not found, so raise exception.
            absolute_ref = self.handle_reference_error(
                orig_ref, self.path(self.group(orig_var))
            )
        else:
            # Found
            logging.info(
                f"            {method} reference to {ref_type} "
                f"{orig_ref!r} resolved as {absolute_ref!r}"
            )

        # If variables refs are limited to coordinate variable,
        # additional check
        if (
            ref_type == "variable"
            and rules.limit_to_scalar_coordinates
            and (
                (
                    "coordinates" not in self.ncattrs(orig_var)
                    or orig_ref not in self.getncattr(orig_var, "coordinates")
                )
                or self._input_ds[absolute_ref].ndim > 0
            )
        ):
            logging.info(
                f"            Reference to {orig_ref!r} is not a "
                "scalar coordinate variable. Assumed to be a standard name."
            )
            absolute_ref = orig_ref

        # Return result
        return absolute_ref

    def search_by_relative_path(self, ref, current_group, search_dim):
        """Search by relative path.

        Resolves the absolute path to a reference within the group
        structure, using search by relative path.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            ref: `str`
                The reference to resolve.

            current_group: `str`
                The current group of the reference.

            search_dim: `bool`
                If True then search for a dimension, otherwise a
                variable.

        :Returns:

            `str`
                The absolute path to the variable.

        """
        # Go up parent groups
        while ref.startswith("../"):
            if current_group.parent is None:
                return None

            ref = ref[3:]
            current_group = current_group.parent

        # Go down child groups
        ref_split = ref.split(group_separator)
        for g in ref_split[:-1]:
            try:
                current_group = current_group.groups[g]
            except KeyError:
                return None

        # Get variable or dimension
        if search_dim:
            elt = current_group.dimensions[ref_split[-1]]
        else:
            elt = current_group.variables[ref_split[-1]]

        # Get absolute reference
        return self.pathname(self.group(elt), self.name(elt))

    def search_by_proximity(
        self,
        ref,
        current_group,
        search_dim,
        local_apex_reached,
        is_coordinate_variable,
    ):
        """Search by proximity.

        Resolves the absolute path to a reference within the group
        structure, using search by proximity.

        First search up in the hierarchy for the reference, until root
        group is reached. If coordinate variable, search until local
        apex is reached, then search down in siblings.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            ref: `str`
                The reference to resolve.

            current_group:
                The current group where searching.

            search_dim: `bool`
                If True then search for a dimension, otherwise a
                variable.

            local_apex_reached: `bool`
                Whether or not the apex has previously been reached.

            is_coordinate_variable: `bool`
                Whether the search is for a coordiante variable.

        :Returns:

            `str` or `None`
                The absolute path to the variable, if found, otherwise
                `None`.

        """
        if search_dim:
            dims_or_vars = current_group.dimensions
        else:
            dims_or_vars = current_group.variables

        # Found in current group
        if ref in dims_or_vars.keys():
            return dims_or_vars[ref]

        local_apex_reached = (
            local_apex_reached or ref in current_group.dimensions.keys()
        )

        # Check if have to continue looking in parent group
        # - normal search: continue until root is reached
        # - coordinate variable: continue until local apex is reached
        if is_coordinate_variable:
            top_reached = local_apex_reached or current_group.parent is None
        else:
            top_reached = current_group.parent is None

        # Search up
        if not top_reached:
            return self.search_by_proximity(
                ref,
                current_group.parent,
                search_dim,
                local_apex_reached,
                is_coordinate_variable,
            )

        elif is_coordinate_variable and local_apex_reached:
            # Coordinate variable and local apex reached, so search
            # down in siblings
            found_elt = None
            for child_group in current_group.groups.values():
                found_elt = self.search_by_proximity(
                    ref,
                    child_group,
                    search_dim,
                    local_apex_reached,
                    is_coordinate_variable,
                )
                if found_elt is not None:
                    break

            return found_elt

        else:
            # Did not find
            return None

    def resolve_references(self, var, old_var):
        """Resolve references.

        In a given variable, replace all references to other variables
        in its attributes by absolute references.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            var:
                The flattened variable in which references should be
                renamed with absolute references, that has the same
                API as `netCDF4.Variable` or `h5netcdf.Variable`.

            old_var:
                The original variable (in group structure), that has
                the same API as `netCDF4.Variable` or
                `h5netcdf.Variable`.

        :Returns:

            `None`

        """
        var_attrs = self.attrs(var)
        for name in referencing_attributes.intersection(var_attrs):
            # Parse attribute value
            parsed_attribute = parse_attribute(name, var_attrs[name])

            # Resolved references in parsed as required by attribute
            # properties
            resolved_parsed_attr = {}

            rules = flattening_rules[name]
            resolve_key = rules.resolve_key
            resolve_value = rules.resolve_value

            for k, v in parsed_attribute.items():
                if resolve_key:
                    k = self.resolve_reference(k, old_var, rules)

                if resolve_value and v is not None:
                    v = [self.resolve_reference(x, old_var, rules) for x in v]

                resolved_parsed_attr[k] = v

            # Re-generate attribute value string with resolved
            # references
            var.setncattr(name, generate_var_attr_str(resolved_parsed_attr))

    def adapt_references(self, var):
        """Adapt references.

        In a given variable, replace all references to variables in
        attributes by references to the new names in the flattened
        netCDF. All references have to be already resolved as absolute
        references.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            var:
                The flattened variable in which references should be
                renamed with new names, that has the same API as
                `netCDF4.Variable` or `h5netcdf.Variable`.

        :Returns:

            `None`

        """
        var_attrs = self.attrs(var)
        for name in referencing_attributes.intersection(var_attrs):
            # Parse attribute value
            value = var_attrs[name]
            parsed_attribute = parse_attribute(name, value)

            adapted_parsed_attr = {}

            rules = flattening_rules[name]
            resolve_key = rules.resolve_key
            resolve_value = rules.resolve_value

            for k, v in parsed_attribute.items():
                if resolve_key:
                    k = self.adapt_name(k, rules)

                if resolve_value and v is not None:
                    v = [self.adapt_name(x, rules) for x in v]

                adapted_parsed_attr[k] = v

            new_attr_value = generate_var_attr_str(adapted_parsed_attr)
            var.setncattr(name, new_attr_value)

            logging.info(
                f"        Value of {self.name(var)}.{name} changed "
                f"from {value!r} to {new_attr_value!r}"
            )

    def adapt_name(self, resolved_ref, rules):
        """Apapt the name.

        Return name of flattened reference. If not found, raise
        exception or continue with a warning.

        .. versionadded:: (cfdm) 1.11.2.0

            resolved_ref: `str`
                The resolved reference.

            rules: `FlatteningRules`
                The flattening rules that apply to the reference.

        :Returns:

            `str`
                The adapted reference.

        """
        # If ref contains Error message, leave as such
        if ref_not_found_error in resolved_ref:
            return resolved_ref

        ref_to_dim = rules.ref_to_dim
        ref_to_var = rules.ref_to_var

        # Select highest priority map
        if ref_to_dim > ref_to_var:
            name_mapping = self._dim_map

        if ref_to_dim < ref_to_var:
            name_mapping = self._var_map

        # Try to find mapping
        try:
            return name_mapping[resolved_ref]

        # If not found, look in other map if allowed
        except KeyError:
            if ref_to_dim and ref_to_var:
                if ref_to_dim < ref_to_var:
                    name_mapping = self._dim_map
                else:
                    name_mapping = self._var_map

                try:
                    return name_mapping[resolved_ref]
                except KeyError:
                    pass

        # If still not found, check if any standard name is allowed
        if rules.accept_standard_names:
            return resolved_ref

        else:
            # If not found, raise exception
            return self.handle_reference_error(resolved_ref)

    def pathname(self, group, name):
        """Compose full path name to an element in a group structure.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            current_group:
                The group containing the dimension or variable.

            name: `str`
                The name of the dimension or variable.

        :Returns:

            `str`
                The absolute path to the dimension or variable

        """
        if self.parent(group) is None:
            return group_separator + name

        return group_separator.join((self.path(group), name))

    def generate_mapping_str(self, input_group, name, new_name):
        """Generate string mapping.

        Generates a string representing the name mapping of an element
        before and after flattening.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            input_group:
                The group containing the non-flattened dimension or
                variable.

            name: `str`
                The name of the non-flattened dimension or variable.

            new_name: `str`
                The name of the flattened dimension or variable.

        :Returns:

            `str`
                A string representing the name mapping for the
                dimension or variable.

        """
        original_pathname = self.pathname(input_group, name)
        return f"{new_name}: {original_pathname}"

    def convert_path_to_valid_name(self, pathname):
        """Generate valid name from path.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            pathname: `str`
                The non-flattened namepath to a dimension or variable.

            new_name: `str`
                A flattened version of *pathname*.
        :Returns:

            `str`
                The valid netCDF name.

        """
        return pathname.replace(group_separator, "", 1).replace(
            group_separator, flattener_separator
        )

    def generate_flattened_name(self, input_group, orig_name):
        """Convert full path of an element to a valid NetCDF name.

        * The name of an element is the concatenation of its
          containing group and its name;

        * replaces ``/`` from paths (forbidden as NetCDF name);

        * if name is longer than 255 characters, replace path to group
          by hash;

        * if name is still too long, replace complete name by hash.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            input_group:
                The group containing the dimension or variable.

            orig_name: `str`
                The original name of the dimension or variable.

        :Returns:

            `str`
                The new valid name of the dimension or variable.

        """
        # If element is at root: no change
        if self.parent(input_group) is None:
            new_name = orig_name

        # If element in child group, concatenate group path and
        # element name
        else:
            full_name = (
                self.convert_path_to_valid_name(self.path(input_group))
                + flattener_separator
                + orig_name
            )
            new_name = full_name

            # If resulting name is too long, hash group path
            if len(new_name) >= max_name_len:
                group_hash = hashlib.sha1(
                    self.path(input_group).encode("UTF-8")
                ).hexdigest()
                new_name = group_hash + flattener_separator + orig_name

                # If resulting name still too long, hash everything
                if len(new_name) >= max_name_len:
                    new_name = hashlib.sha1(
                        full_name.encode("UTF-8")
                    ).hexdigest()

        return new_name

    def handle_reference_error(self, ref, context=None):
        """Handle reference error.

        Depending on the `_strict` mode, either raise an exception or
        log a warning. If not strict then a reference placeholder is
        returned.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            ref: `str`
                The reference

            context: `str`
                Additional context information to add to message.

        :Returns:

            `str`
                The error message, or if `_strict` is `True` then an
                `UnresolvedReferenceException` is raised.

        """
        message = f"Reference {ref!r} could not be resolved"
        if context is not None:
            message = f"{message} from {context}"

        if self._strict:
            raise UnresolvedReferenceException(message)

        warnings.warn(message)
        return f"{ref_not_found_error}_{ref}"


class AttributeParsingException(Exception):
    """Exception for unparsable attribute.

    .. versionadded:: (cfdm) 1.11.2.0

    """

    pass


class UnresolvedReferenceException(Exception):
    """Exception for unresolvable references in attributes.

    .. versionadded:: (cfdm) 1.11.2.0

    """

    pass
